import pytest

from refinitiv.data import _configure as configure
from .conftest import ENV_NAME, ENV_VALUE


def test_substitution_path(tmp_path):
    import os
    import platform

    env = os.environ
    p = tmp_path / "config.json"
    subs_name = ""
    if platform.system() == "Linux":
        subs_name = "PATH"
    elif platform.system() == "Windows":
        subs_name = "Path"
    else:
        subs_name = "PATH"
    p.write_text('{"prop": "${%s}"}' % subs_name)

    configure.reload()
    cfg = configure._read_config_file(p)
    assert cfg.get("prop") == env.get(subs_name)


def test_substitution_no(tmp_path):
    p = tmp_path / "config.json"
    p.write_text('{"prop": "value"}')

    configure.reload()
    cfg = configure._read_config_file(p)
    assert cfg.get("prop") == "value"


def test_substitution_many(tmp_path, monkeypatch):
    monkeypatch.setenv("TEST_NAME_1", "TEST_VALUE_1")
    monkeypatch.setenv("TEST_NAME_2", "TEST_VALUE_2")
    p = tmp_path / "config.json"
    p.write_text('{"prop_1": "${TEST_NAME_1}", "prop_2": "${TEST_NAME_2}"}')

    configure.reload()
    cfg = configure._read_config_file(p)
    assert cfg.get("prop_1") == "TEST_VALUE_1"
    assert cfg.get("prop_2") == "TEST_VALUE_2"


def test_substitution_empty(tmp_path):
    p = tmp_path / "config.json"
    p.write_text('{"prop": "${}"}')

    configure.reload()
    cfg = configure._read_config_file(p)
    assert cfg.get("prop") == "${}"


@pytest.mark.parametrize("value", [r"win\path\to\test", r"nix/path/to/test"])
def test_substitution_platform_path(tmp_path, monkeypatch, value):
    monkeypatch.setenv("TEST_PATH", value)
    p = tmp_path / "config.json"
    p.write_text('{"prop": "${TEST_PATH}", "prop_1": "path\\to\\test"}')

    configure.reload()
    cfg = configure._read_config_file(p)
    assert cfg.get("prop") == value


def test_substitution_exists(mock_env_user, tmp_path):
    p = tmp_path / "config.json"
    p.write_text('{"prop": "${%s}"}' % ENV_NAME)
    configure.reload()
    cfg = configure._read_config_file(p)
    assert cfg.get("prop") == ENV_VALUE


def test_substitution_not_exists(mock_env_missing, tmp_path):
    p = tmp_path / "config.json"
    p.write_text('{"prop": "${%s}"}' % ENV_NAME)
    configure.reload()
    cfg = configure._read_config_file(p)
    assert cfg.get("prop") == ENV_NAME


@pytest.mark.parametrize(
    "testing_data",
    [
        ({"root": ["root_value"], "content": {"prop": "${root}"}}, "${root}"),
        ({"root": ["root_value"], "content": {"prop": "${root:0}"}}, "root_value"),
        (
            {"root": ["foo", "root_value"], "content": {"prop": "${root:1}"}},
            "root_value",
        ),
        (
            {"root": [{"foo": "root_value"}], "content": {"prop": "${root:0:foo}"}},
            "root_value",
        ),
        ({"root": "root_value", "content": {"prop": "${root}"}}, "root_value"),
        ({"ro-ot": "root_value", "content": {"prop": "${ro-ot}"}}, "root_value"),
        ({"ro_ot": "root_value", "content": {"prop": "${ro_ot}"}}, "root_value"),
        (
            {"root": "root_value", "content": {"prop": "${root}${root}"}},
            "root_valueroot_value",
        ),
        (
            {"root": {"prop": "root_value"}, "content": {"prop": "${root:prop}"}},
            "root_value",
        ),
        (
            {
                "root": {"prop": "root_value"},
                "content": {"prop": "${root:prop}${root:prop}"},
            },
            "root_valueroot_value",
        ),
        (
            {
                "root": {"prop_1": "root_1_value", "prop_2": "root_2_value"},
                "content": {"prop": "${root:prop_1}${root:prop_2}"},
            },
            "root_1_valueroot_2_value",
        ),
        (
            {"path": {"to": {"prop": "value"}}, "content": {"prop": "${path:to:prop}"}},
            "value",
        ),
        (
            {
                "path": {"to": {"prop": "value"}},
                "content": {"prop": "${path:to:prop}_1"},
            },
            "value_1",
        ),
        (
            {
                "path": {"to": {"prop": "value"}},
                "content": {"prop_": "value_", "prop": "${content:prop_}"},
            },
            "value_",
        ),
    ],
)
def test_substitute_values(testing_data):
    input_data, expected = testing_data
    configure.reload()
    cfg = configure._substitute_values(input_data, input_data)
    assert cfg.get("content").get("prop") == expected, input_data


@pytest.mark.parametrize(
    "testing_data",
    [
        ({"content": {"prop": "${root}"}}, "${root}"),
        ({"content": {"prop": "${root}${root}"}}, "${root}${root}"),
    ],
)
def test_substitute_values_cannot_substitute(testing_data):
    input_data, expected = testing_data
    configure.reload()
    cfg = configure._substitute_values(input_data, input_data)
    assert cfg.get("content").get("prop") == expected, input_data


def test_python_configuration_package():
    """
    https://jira.refinitiv.com/browse/EAPI-1777
    https://github.com/tr11/python-configuration/issues/49
    """

    from refinitiv.data._external_libraries import (
        python_configuration as ext_config_mod,
    )

    d = {"path": {"to": {"value-a": "A", "value-b": "B"}}}
    base_cfg = ext_config_mod.config_from_dict(d)

    d = {"path": {"to": {"value-a": "C"}}}
    cfg = ext_config_mod.config_from_dict(d)

    cfg_set = ext_config_mod.ConfigurationSet(cfg, base_cfg)

    path_config = cfg_set.get("path")
    assert path_config == {"to.value-a": "C", "to.value-b": "B"}



from refinitiv.data.content.ipa.curves import zc_curve_definitions
from tests.unit.conftest import remove_dunder_methods, remove_private_attributes


def test_zc_curve_definitions_scope():
    expected_attributes = ["AssetClass", "Definition", "Definitions", "RiskType"]
    testing_attributes = dir(zc_curve_definitions)
    testing_attributes = remove_dunder_methods(testing_attributes)
    testing_attributes = remove_private_attributes(testing_attributes)
    assert expected_attributes == testing_attributes



# coding: utf8


from typing import Optional

from ..._object_definition import ObjectDefinition
from ._enums import PriceSide


class PricingParameters(ObjectDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    price_side : PriceSide, optional
        The quoted price side of the instrument. optional. default value is 'mid'.
    exercise_date : str, optional

    market_data_date : str, optional
        The date at which the market data is retrieved. the value is expressed in iso
        8601 format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z'). it
        should be less or equal tovaluationdate). optional. by
        default,marketdatadateisvaluationdateor today.
    market_value_in_deal_ccy : float, optional
        The market value of the instrument. the value is expressed in the deal currency.
        optional. no default value applies. note that premium takes priority over
        volatility input.
    nb_iterations : int, optional
        The number of steps for the bermudan swaption pricing via the hull-white
        one-factor (hw1f) tree.  no default value applies.
    report_ccy : str, optional
        The reporting currency code, expressed in iso 4217 alphabetical format (e.g.,
        'usd'). it is set for the fields ending with 'xxxinreportccy'. optional. the
        default value is the notional currency.
    simulate_exercise : bool, optional
        Tells if in case of future cashflows should be considered as exercised or not.
        possible values:    true,    false.
    valuation_date : str, optional
        The valuation date for pricing. If not set the valuation date is equal to
        market_data_date or Today. For assets that contains a settlementConvention,
        the default valuation date  is equal to the settlementdate of the Asset that is
        usually the TradeDate+SettlementConvention.

    Examples
    --------
     >>> import refinitiv.data.content.ipa.financial_contracts as rdf
     >>> rdf.swaption.PricingParameters(valuation_date="2020-04-24", nb_iterations=80)
    """

    def __init__(
        self,
        price_side: Optional[PriceSide] = None,
        exercise_date: Optional[str] = None,
        market_data_date: Optional[str] = None,
        market_value_in_deal_ccy: Optional[float] = None,
        nb_iterations: Optional[int] = None,
        report_ccy: Optional[str] = None,
        simulate_exercise: Optional[bool] = None,
        valuation_date: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.price_side = price_side
        self.exercise_date = exercise_date
        self.market_data_date = market_data_date
        self.market_value_in_deal_ccy = market_value_in_deal_ccy
        self.nb_iterations = nb_iterations
        self.report_ccy = report_ccy
        self.simulate_exercise = simulate_exercise
        self.valuation_date = valuation_date

    @property
    def price_side(self):
        """
        The quoted price side of the instrument. optional. default value is 'mid'.
        :return: enum PriceSide
        """
        return self._get_enum_parameter(PriceSide, "priceSide")

    @price_side.setter
    def price_side(self, value):
        self._set_enum_parameter(PriceSide, "priceSide", value)

    @property
    def exercise_date(self):
        """
        :return: str
        """
        return self._get_parameter("exerciseDate")

    @exercise_date.setter
    def exercise_date(self, value):
        self._set_parameter("exerciseDate", value)

    @property
    def market_data_date(self):
        """
        The market data date for pricing.
        By default, the marketDataDate date is the ValuationDate or Today.
        :return: str
        """
        return self._get_parameter("marketDataDate")

    @market_data_date.setter
    def market_data_date(self, value):
        self._set_parameter("marketDataDate", value)

    @property
    def market_value_in_deal_ccy(self):
        """
        MarketValueInDealCcy to override and that will be used as pricing analysis input to compute VolatilityPercent.
        No override is applied by default. Note that Premium takes priority over Volatility input.
        :return: float
        """
        return self._get_parameter("marketValueInDealCcy")

    @market_value_in_deal_ccy.setter
    def market_value_in_deal_ccy(self, value):
        self._set_parameter("marketValueInDealCcy", value)

    @property
    def nb_iterations(self):
        """
        Used for Bermudans and HW1F tree.
        :return: int
        """
        return self._get_parameter("nbIterations")

    @nb_iterations.setter
    def nb_iterations(self, value):
        self._set_parameter("nbIterations", value)

    @property
    def report_ccy(self):
        """
        The reporting currency code, expressed in iso 4217 alphabetical format (e.g.,
        'usd'). it is set for the fields ending with 'xxxinreportccy'. optional. the
        default value is the notional currency.
        :return: str
        """
        return self._get_parameter("reportCcy")

    @report_ccy.setter
    def report_ccy(self, value):
        self._set_parameter("reportCcy", value)

    @property
    def simulate_exercise(self):
        """
        Tells if in case of future cashflows should be considered as exercised or not.
        possible values:    true,    false.
        :return: bool
        """
        return self._get_parameter("simulateExercise")

    @simulate_exercise.setter
    def simulate_exercise(self, value):
        self._set_parameter("simulateExercise", value)

    @property
    def valuation_date(self):
        """
        The valuation date for pricing.
        If not set the valuation date is equal to MarketDataDate or Today.
        For assets that contains a settlementConvention, the default valuation date  is equal to
        the settlementdate of the Asset that is usually the TradeDate+SettlementConvention.
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)



"""Configuration class."""

import base64
from contextlib import contextmanager
from copy import deepcopy
from sys import version_info
from typing import (
    Any,
    Dict,
    ItemsView,
    Iterator,
    KeysView,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
    ValuesView,
    cast,
)

from .helpers import (
    AttributeDict,
    InterpolateEnumType,
    InterpolateType,
    as_bool,
    clean,
    interpolate_object,
)

if version_info < (3, 8):  # pragma: no cover
    from collections import OrderedDict


class Configuration:
    """
    Configuration class.

    The Configuration class takes a dictionary input with keys such as

        - ``a1.b1.c1``
        - ``a1.b1.c2``
        - ``a1.b2.c1``
        - ``a1.b2.c2``
        - ``a2.b1.c1``
        - ``a2.b1.c2``
        - ``a2.b2.c1``
        - ``a2.b2.c2``
    """

    def __init__(
        self,
        config_: Mapping[str, Any],
        lowercase_keys: bool = False,
        interpolate: InterpolateType = False,
        interpolate_type: InterpolateEnumType = InterpolateEnumType.STANDARD,
    ):
        """
        Constructor.

        :param config_: a mapping of configuration values. Keys need to be strings.
        :param lowercase_keys: whether to convert every key to lower case.
        """
        self._lowercase = lowercase_keys
        self._interpolate = {} if interpolate is True else interpolate
        self._interpolate_type = interpolate_type
        self._config: Dict[str, Any] = self._flatten_dict(config_)
        self._default_levels: Optional[int] = 1

    def __eq__(self, other):  # type: ignore
        """Equality operator."""
        return self.as_dict() == Configuration(other).as_dict()

    def _filter_dict(self, d: Dict[str, Any], prefix: str) -> Dict[str, Any]:
        """
        Filter a dictionary and return the items that are prefixed by :attr:`prefix`.

        :param d: dictionary
        :param prefix: prefix to filter on
        """
        if self._lowercase:
            return {
                k[(len(prefix) + 1) :].lower(): v
                for k, v in d.items()
                for k, v in d.items()
                if k.startswith(prefix + ".")
            }
        else:
            return {
                k[(len(prefix) + 1) :]: v
                for k, v in d.items()
                if k.startswith(prefix + ".")
            }

    def _flatten_dict(self, d: Mapping[str, Any]) -> Dict[str, Any]:
        """
        Flatten one level of a dictionary.

        :param d: dict
        :return: a flattened dict
        """
        nested = {k for k, v in d.items() if isinstance(v, (dict, Configuration))}
        if self._lowercase:
            result = {
                k.lower() + "." + ki: vi
                for k in nested
                for ki, vi in self._flatten_dict(d[k]).items()
            }
            result.update(
                (k.lower(), v)
                for k, v in d.items()
                if not isinstance(v, (dict, Configuration))
            )
        else:
            result = {
                k + "." + ki: vi
                for k in nested
                for ki, vi in self._flatten_dict(d[k]).items()
            }
            result.update(
                (k, v) for k, v in d.items() if not isinstance(v, (dict, Configuration))
            )
        return result

    def _get_subset(self, prefix: str) -> Union[Dict[str, Any], Any]:
        """
        Return the subset of the config dictionary whose keys start with :attr:`prefix`.

        :param prefix: string
        :return: dict
        """
        d = {
            k[(len(prefix) + 1) :]: v
            for k, v in self._config.items()
            if k.startswith(prefix + ".")
        }
        if not d:
            prefixes = prefix.split(".")
            if len(prefixes) == 1:
                return deepcopy(self._config.get(prefix, {}))
            d = self._config
            while prefixes:  # pragma: no branches
                p = prefixes[0]
                new_d = self._filter_dict(d, p)
                if new_d == {}:
                    return deepcopy(d.get(p, {}) if len(prefixes) == 1 else {})
                d = new_d
                prefixes = prefixes[1:]
        return deepcopy(d)

    def __getitem__(self, item: str) -> Union["Configuration", Any]:  # noqa: D105
        v = self._get_subset(item)

        if v == {}:
            raise KeyError(item)
        if isinstance(v, dict):
            return Configuration(v)
        elif self._interpolate is not False:
            d = self.as_dict()
            d.update(cast(Dict[str, str], self._interpolate))
            return interpolate_object(item, v, [d], self._interpolate_type)
        else:
            return v

    def __getattr__(self, item: str) -> Any:  # noqa: D105
        try:
            return self[item]
        except KeyError:
            raise AttributeError(item)

    def get(self, key: str, default: Any = None) -> Union[dict, Any]:
        """
        Get the configuration values corresponding to :attr:`key`.

        :param key: key to retrieve
        :param default: default value in case the key is missing
        :return: the value found or a default
        """
        return self.as_dict().get(key, default)

    def as_dict(self) -> dict:
        """Return the representation as a dictionary."""
        return self._config

    def as_attrdict(self) -> AttributeDict:
        """Return the representation as an attribute dictionary."""
        return AttributeDict(
            {
                x: Configuration(v).as_attrdict() if isinstance(v, dict) else v
                for x, v in self.items(levels=1)
            }
        )

    def get_bool(self, item: str) -> bool:
        """
        Get the item value as a bool.

        :param item: key
        """
        return as_bool(self[item])

    def get_str(self, item: str, fmt: str = "{}") -> str:
        """
        Get the item value as an int.

        :param item: key
        :param fmt: format to use
        """
        return fmt.format(self[item])

    def get_int(self, item: str) -> int:
        """
        Get the item value as an int.

        :param item: key
        """
        return int(self[item])

    def get_float(self, item: str) -> float:
        """
        Get the item value as a float.

        :param item: key
        """
        return float(self[item])

    def get_list(self, item: str) -> List[Any]:
        """
        Get the item value as a list.

        :param item: key
        """
        return list(self[item])

    def get_dict(self, item: str) -> dict:
        """
        Get the item values as a dictionary.

        :param item: key
        """
        return dict(self._get_subset(item))

    def base64encode(self, item: str) -> bytes:
        """
        Get the item value as a Base64 encoded bytes instance.

        :param item: key
        """
        b = self[item]
        b = b if isinstance(b, bytes) else b.encode()
        return base64.b64encode(b)

    def base64decode(self, item: str) -> bytes:
        """
        Get the item value as a Base64 decoded bytes instance.

        :param item: key
        """
        b = self[item]
        b = b if isinstance(b, bytes) else b.encode()
        return base64.b64decode(b, validate=True)

    def keys(
        self, levels: Optional[int] = None
    ) -> Union["Configuration", Any, KeysView[str]]:
        """Return a set-like object providing a view on the configuration keys."""
        assert levels is None or levels > 0
        levels = self._default_levels if levels is None else levels
        try:
            return self["keys"]  # don't filter levels, existing attribute
        except KeyError:
            return cast(
                KeysView[str],
                list(
                    {
                        ".".join(x.split(".")[:levels])
                        for x in set(self.as_dict().keys())
                    }
                ),
            )

    def values(
        self, levels: Optional[int] = None
    ) -> Union["Configuration", Any, ValuesView[Any]]:
        """Return a set-like object providing a view on the configuration values."""
        assert levels is None or levels > 0
        levels = self._default_levels if levels is None else levels
        try:
            return self["values"]
        except KeyError:
            return dict(self.items(levels=levels)).values()

    def items(
        self, levels: Optional[int] = None
    ) -> Union["Configuration", Any, ItemsView[str, Any]]:
        """Return a set-like object providing a view on the configuration items."""
        assert levels is None or levels > 0
        levels = self._default_levels if levels is None else levels
        try:
            return self["items"]
        except KeyError:
            keys = cast(KeysView[str], self.keys(levels=levels))
            return {k: self._get_subset(k) for k in keys}.items()

    def __iter__(self) -> Iterator[Tuple[str, Any]]:  # noqa: D105
        return iter(dict(self.items()))  # type: ignore

    def __reversed__(self) -> Iterator[Tuple[str, Any]]:  # noqa: D105
        if version_info < (3, 8):
            return OrderedDict(  # type: ignore
                reversed(list(self.items()))
            )  # pragma: no cover
        else:
            return reversed(dict(self.items()))  # type: ignore

    def __len__(self) -> int:  # noqa: D105
        return len(self.keys())

    def __setitem__(self, key: str, value: Any) -> None:  # noqa: D105
        self.update({key: value})

    def __delitem__(self, prefix: str) -> None:  # noqa: D105
        """
        Filter a dictionary and delete the items that are prefixed by :attr:`prefix`.

        :param prefix: prefix to filter on to delete keys
        """
        remove = []
        for k in self._config:
            kl = k.lower() if self._lowercase else k
            if kl == prefix or kl.startswith(prefix + "."):
                remove.append(k)
        if not remove:
            raise KeyError("No key with prefix '%s' found." % prefix)
        for k in remove:
            del self._config[k]

    def __contains__(self, prefix: str) -> bool:  # noqa: D105
        try:
            self[prefix]
            return True
        except KeyError:
            return False

    def clear(self) -> None:
        """Remove all items."""
        self._config.clear()

    def copy(self) -> "Configuration":
        """Return shallow copy."""
        return Configuration(self._config)

    def pop(self, prefix: str, value: Any = None) -> Any:  # NOSONAR
        """
        Remove keys with the specified prefix and return the corresponding value.

        If the prefix is not found a KeyError is raised.
        """
        try:
            value = self[prefix]
            del self[prefix]
        except KeyError:
            if value is None:
                raise
        return value

    def setdefault(self, key: str, default: Any = None) -> Any:
        """
        Insert key with a value of default if key is not in the Configuration.

        Return the value for key if key is in the Configuration, else default.
        """
        try:
            return self[key]
        except KeyError:
            self[key] = default
        return self[key]

    def update(self, other: Mapping[str, Any]) -> None:
        """Update the Configuration with another Configuration object or Mapping."""
        self._config.update(self._flatten_dict(other))

    def reload(self) -> None:  # pragma: no cover
        """
        Reload the configuration.

        This method is not implemented for simple Configuration objects and is
        intended only to be used in subclasses.
        """
        raise NotImplementedError()

    @contextmanager
    def dotted_iter(self) -> Iterator["Configuration"]:
        """
        Context manager for dotted iteration.

        This context manager changes all the iterator-related functions
        to include every nested (dotted) key instead of just the top level.
        """
        self._default_levels = None
        try:
            yield self
        finally:
            self._default_levels = 1

    def __repr__(self) -> str:  # noqa: D105
        return "<Configuration: %s>" % hex(id(self))

    def __str__(self) -> str:  # noqa: D105
        return str({k: clean(k, v) for k, v in sorted(self.as_dict().items())})



from typing import TYPE_CHECKING

from ._bid_ask import BidAsk

if TYPE_CHECKING:
    from .._types import OptStr, OptFloat


class FxSpotQuote(BidAsk):
    """
    An object that describes a fx spot quote

    Parameters
    ----------
    source : str, optional
        Contributor id (cid)
    ask : float, optional
        Ask value
    bid : float, optional
        Bid value
    """

    def __init__(
        self,
        source: "OptStr" = None,
        ask: "OptFloat" = None,
        bid: "OptFloat" = None,
    ) -> None:
        super().__init__(ask, bid)
        self.source = source

    @property
    def source(self):
        return self._get_param("source")

    @source.setter
    def source(self, value):
        self._set_param("source", value)



from refinitiv.data.delivery import rdp_stream
from refinitiv.data.delivery._stream import _RDPStream, StreamState
from tests.unit.conftest import StubSession


def test_rdp_stream_does_not_blocking_by_alarm_event():
    # given
    session = StubSession()
    rdp_stream = _RDPStream(1, session, "name", None, None, None, None)
    rdp_stream._state = StreamState.Opening

    # when
    args = {"state": {}}
    rdp_stream._do_on_stream_alarm({}, args)

    # then
    assert rdp_stream._opened.is_set()


def test_rdp_stream_doesnot_change_universe_as_dict():
    # given
    input_universe = {
        "instrumentType": "FxCross",
        "instrumentDefinition": {
            "instrumentTag": "USDAUD",
            "fxCrossType": "FxSpot",
            "fxCrossCode": "USDAUD",
        },
    }
    expected_universe = input_universe.copy()

    session = StubSession()

    # when
    definition = rdp_stream.Definition(
        service=None,
        api="streaming.quantitative-analytics.endpoints.financial-contracts",
        universe=input_universe,
        parameters=None,
        view=[
            "InstrumentTag",
            "FxSpot_BidMidAsk",
            "ErrorCode",
            "Ccy1SpotDate",
            "Ccy2SpotDate",
        ],
    )
    stream = definition.get_stream(session)

    # then
    assert definition._universe == expected_universe
    assert stream._universe == expected_universe



import gzip
import os

from ._tools import path_join, create_dirs_if_no_exists
from ._tools import remove_ext
from ..._core.log_reporter import _LogReporter


class _Unpacker(_LogReporter):
    def __init__(self, logger) -> None:
        super().__init__(logger=logger)

    def unpack(self, filepath, extract_path="") -> str:
        self._debug("start extracting file")

        *_, filename_ext = os.path.split(filepath)
        extracted_filename = remove_ext(filename_ext)

        if not extract_path:
            extract_filepath = remove_ext(filepath)

        else:
            create_dirs_if_no_exists(extract_path)
            extract_filepath = path_join(extract_path, extracted_filename)

        self._debug(f"extract file '{filename_ext}' to '{extract_filepath}'")

        with gzip.open(filepath, "rb") as gzip_file:
            try:
                with open(extract_filepath, "wb+") as file:
                    file.write(gzip_file.read())
            except Exception as e:
                self._error(str(e))
                raise e

        self._debug("file extracted successfully")

        return extract_filepath



from typing import List, Optional

from ..._types import OptDateTime
from ...content.ipa.dates_and_calendars.is_working_day import Definition


def is_working_day(
    date: "OptDateTime" = None,
    currencies: Optional[List[str]] = None,
    calendars: Optional[List[str]] = None,
) -> bool:
    """
    With this function you can get check if date is
    working day or not for any of supplied calendars or currencies.

    Parameters
    ----------
        date: str or datetime or timedelta, optional
            Date to test.
        calendars: list of str, optional
            Calendars to use the date for working day or weekend.
            Optional if currencies is provided.
        currencies: list of str, optional
            Currencies to use the date for working day or weekend.
            Optional if calendars is provided.

    Returns
    -------
    bool
        If requested day is working day returns True, otherwise returns False

    Examples
    --------
     >>> import datetime
     >>> import refinitiv.data as rd
     >>>
     >>> rd.open_session("platform.default")
     >>>
     >>> is_working = rd.dates_and_calendars.is_working_day(
     ...    date=datetime.datetime(2020, 7, 10),
     ...    currencies=["EUR"]
     >>>)
    """

    response = Definition(
        date=date, calendars=calendars, currencies=currencies
    ).get_data()

    return response.data.day.is_working_day



from types import SimpleNamespace

from refinitiv.data._config_defaults import data as _config_defaults_data
from refinitiv.data._external_libraries import python_configuration as ext_config_mod
from refinitiv.data.content.esg.bulk._db_manager import DBManager
from tests.unit.conftest import StubLogger


def test_build_df():
    connection = SimpleNamespace()
    cursor = SimpleNamespace()
    cursor.description = []
    connection.cursor = lambda: cursor
    config = ext_config_mod.config_from_dict(
        _config_defaults_data["bulk"]["esg"]["standard_scores"]["db"]
    )
    actions = SimpleNamespace()
    db_manager = DBManager(connection, config, actions, StubLogger())
    db_manager.exec = lambda *args: {}

    data = db_manager.get_data(universe=[], fields=[])

    assert data.df.empty


def test_build_df_when_exec_raised_error():
    def exec(*args, **kwargs):
        raise Exception()

    connection = SimpleNamespace()
    cursor = SimpleNamespace()
    cursor.description = []
    connection.cursor = lambda: cursor
    config = ext_config_mod.config_from_dict(
        _config_defaults_data["bulk"]["esg"]["standard_scores"]["db"]
    )
    actions = SimpleNamespace()
    actions.add = lambda *args, **kwargs: None
    db_manager = DBManager(connection, config, actions, StubLogger())
    db_manager.exec = exec

    data = db_manager.get_data(universe=[], fields=[])

    assert data.df.empty



import pytest

from tests.unit.conftest import remove_dunder_methods, remove_private_attributes
import refinitiv.data.content.ipa as ipa


def test_surface_output():
    with pytest.raises(AttributeError):
        _ = ipa.SurfaceOutput


def test_fx_surface_scope():
    with pytest.raises(AttributeError):
        _ = ipa.Definition
    with pytest.raises(AttributeError):
        _ = ipa.PricingParameters

    with pytest.raises(AttributeError):
        _ = ipa.surface.Definition
    with pytest.raises(AttributeError):
        _ = ipa.surface.PricingParameters


def test_eti_surface_scope():
    with pytest.raises(AttributeError):
        _ = ipa.Definition
    with pytest.raises(AttributeError):
        _ = ipa.PricingParameters

    with pytest.raises(AttributeError):
        _ = ipa.surface.Definition
    with pytest.raises(AttributeError):
        _ = ipa.surface.PricingParameters


def test_ir_surface_scope():
    with pytest.raises(AttributeError):
        _ = ipa.Definition
    with pytest.raises(AttributeError):
        _ = ipa.PricingParameters

    with pytest.raises(AttributeError):
        _ = ipa.surface.Definition
    with pytest.raises(AttributeError):
        _ = ipa.surface.PricingParameters



import allure
import pytest

from refinitiv.data.content.ipa.surfaces import swaption
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ipa.conftest import (
    check_surface,
    check_surface_curve,
    check_surface_point,
)
from tests.integration.content.ipa.surfaces.surface_swaption.conftest import (
    surface_swaption_definition,
    invalid_surface_swaption_definition,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_response_status,
    check_non_empty_response_data,
)


@allure.suite("Content object - Surface Swaption")
@allure.feature("Content object - Surface Swaption")
@allure.severity(allure.severity_level.CRITICAL)
class TestSurfaceSwaption:
    @pytest.mark.xfail(reason="https://jira.refinitiv.com/browse/EAPI-3135")
    @allure.title(
        "Create a surface swaption definition object with valid params and get data"
    )
    @pytest.mark.caseid("37819904")
    @pytest.mark.smoke
    def test_surface_swaption_with_valid_params_and_get_data(self, open_session):
        response = surface_swaption_definition().get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        surface = response.data.surface
        check_surface(surface)
        curve = surface.get_curve("0.25", swaption.Axis.X)
        check_surface_curve(curve)
        point = surface.get_point("0.25", "1Y")
        check_surface_point(point)

    @pytest.mark.xfail(reason="https://jira.refinitiv.com/browse/EAPI-3135")
    @allure.title(
        "Create a surface swaption definition object with valid params and get data asynchronously"
    )
    @pytest.mark.caseid("37819906")
    async def test_surface_swaption_with_valid_params_and_get_data_async(
        self, open_session_async
    ):
        valid_response, invalid_response = await get_async_response_from_definitions(
            surface_swaption_definition(), invalid_surface_swaption_definition()
        )
        check_response_status(valid_response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(valid_response)
        surface = valid_response.data.surface
        check_surface(surface)
        curve = surface.get_curve("0.5", swaption.Axis.X)
        check_surface_curve(curve)
        point = surface.get_point("0.5", "2Y")
        check_surface_point(point)

        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="QPS-Pricer.1923",
            expected_error_message="Invalid market data input for IR volatility. no swaption vol source for INVAL",
        )

    @pytest.mark.xfail(reason="https://jira.refinitiv.com/browse/EAPI-4119")
    @allure.title("Create a surface swaption - short definition")
    @pytest.mark.caseid("37819907")
    def test_surface_swaption_with_simple_definition(self, open_session):
        response = swaption.Definition(
            underlying_definition={"instrumentCode": "EUR"}
        ).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        surface = response.data.surface
        check_surface(surface)
        curve = surface.get_curve("1.5", "X")
        check_surface_curve(curve)
        point = surface.get_point("1.5", "3Y")
        check_surface_point(point)

    @allure.title("Create a invalid surface swaption definition and get RDError")
    @pytest.mark.caseid("39606893")
    def test_invalid_surface_swaption_definition(self, open_session):
        with pytest.raises(RDError) as error:
            invalid_surface_swaption_definition().get_data()
        assert (
            str(error.value)
            == "Error code VolSurf.10300 | The service failed to build the volatility surface. (Additional technical information: Analysis error : pricing item is invalid. SABR volatility cube is not available for 'INVAL', as it has no suitable cap volatility surface provider for caplets vol stripping.)"
        )



from typing import Optional, List

from ._create._curve_definition_description import (
    CrossCurrencyCurveDefinitionDescription,
)
from ._instruments_segment import CrossCurrencyInstrumentsSegment
from ._override_bid_ask import OverrideBidAsk
from ._override_fx_forward_turn import OverrideFxForwardTurn
from ._update._curve_update_definition import (
    CrossCurrencyCurveUpdateDefinition,
)


CurveCreateDefinition = CrossCurrencyCurveDefinitionDescription
CurveUpdateDefinition = CrossCurrencyCurveUpdateDefinition
OptOverrides = Optional[List[OverrideBidAsk]]
Segments = List[CrossCurrencyInstrumentsSegment]
OptTurns = Optional[List[OverrideFxForwardTurn]]



"""Module with 'in-memory Database' based on pandas, with several tables.

Goal is to replace discovery search templates API in target templates testing without
need to emulate it, patch objects, etc..
Almost everything can be tested that way. Because all logic of TargetTemplate is not
tied to search template API.
"""

from typing import Optional

import pandas as pd

from refinitiv.data.discovery._search_templates.base import Target

__all__ = ["CatsDatabaseTarget"]

DB = {
    "cats": pd.DataFrame(
        columns=("name", "age", "gender", "breed", "shelter"),
        data=[
            ("Richard", 12, "male", "Birman", "01"),
            ("Bear", 9, "male", "Sphynx", "02"),
            ("Thomas", 25, "male", "Birman", "01"),
            ("Maverick", 19, "male", "Munchkin", "03"),
            ("Einstein", 21, "male", "Ragdoll", "03"),
            ("Athena", 8, "female", "Sphynx", "02"),
            ("Rita", 12, "female", "Birman", "01"),
            ("Luna", 16, "female", "Ragdoll", "01"),
            ("Elsa", 32, "female", "Munchkin", "02"),
            ("Maeve", 19, "female", "Bombay", "02"),
        ],
    ),
    "breeds": pd.DataFrame(
        columns=("name", "origin", "body_type", "coat"),
        data=[
            ("Birman", "France", "cobby", "semi-long"),
            ("Sphynx", "Canada,Europe", "oriental", "hairless"),
            ("Bombay", "US,Burma", "cobby", "short"),
            ("Munchkin", "US", "dwarf", "short,long"),
            ("Ragdoll", "US", "cobby", "long"),
        ],
    ),
    "shelters": pd.DataFrame(
        columns=("name", "city"),
        data=[
            ("01", "Paris"),
            ("02", "Chartres"),
            ("03", "Paris"),
        ],
    ),
}


class CatsDatabaseTarget(Target):
    """Toy 'Database' target for the local testing"""

    def __init__(self):
        super().__init__()
        self.args_names = {"view", "limit", "filter"}

    def __call__(
        self,
        view: str,
        limit: int = 3,
        filter: Optional[str] = None,
    ):
        """

        Parameters
        ----------
        view: Literal["cats", "breeds", "shelters"]
            "Table" to select data from
        limit: int
            Maximum number of results
        filter: str
            Filter expression that will be passed to DataFrame.query

        Returns
        -------
        Dataframe
        """
        result = DB[view].copy()
        if filter is not None:
            result.query(filter, inplace=True)
        return result[:limit]



from enum import Enum, unique


@unique
class IssuerType(Enum):
    AGENCY = "Agency"
    CORPORATE = "Corporate"
    MUNIS = "Munis"
    NON_FINANCIALS = "NonFinancials"
    SOVEREIGN = "Sovereign"
    SUPRANATIONAL = "Supranational"



# coding: utf8

from enum import Enum, unique


@unique
class BarrierMode(Enum):
    AMERICAN = "American"
    EARLY_END_WINDOW = "EarlyEndWindow"
    EUROPEAN = "European"
    FORWARD_START_WINDOW = "ForwardStartWindow"
    UNDEFINED = "Undefined"



def has_property_names_in_class(cls, expected_property_names):
    if isinstance(expected_property_names, dict):
        expected_property_names = list(expected_property_names.keys())
    return set(get_property_names(cls)) == set(expected_property_names)


def get_property_names(cls):
    return [p for p in dir(cls) if isinstance(getattr(cls, p, None), property)]



from refinitiv.data._tools._converter import (
    try_copy_to_list,
    convert_dict_to_df,
    convert_content_data_to_df,
)
import pytest


def test_convert_dict_to_df_empty_data():
    # given
    data = []
    columns = ["a", "b"]

    # when
    df1 = convert_dict_to_df(data, columns)
    df2 = convert_dict_to_df(data, [])

    # then
    assert list(df1) == columns
    assert df1.empty
    assert df2.empty


def test_convert_dict_to_df_with_data():
    # given
    data = [[1, 2], [3, 4]]
    columns = ["a", "b"]

    # when
    df = convert_dict_to_df(data, columns)

    # then
    assert list(df) == columns
    assert df.size == 4
    assert (df.get("a") == [1, 3]).all()
    assert (df.get("b") == [2, 4]).all()


def test_content_data_to_df_empty():
    # given
    data = {}

    # when
    df = convert_content_data_to_df(data)

    # then
    assert df.empty


def test_content_data_to_df():
    # given
    data = {
        "headers": [{"name": "a", "title": "A1"}, {"name": "b", "title": "B1"}],
        "data": [[1, 2], [3, 4]],
    }

    # when
    df_name_as_headers = convert_content_data_to_df(
        data, use_field_names_in_headers=True
    )
    df_title_as_headers = convert_content_data_to_df(
        data, use_field_names_in_headers=False
    )

    df_default_headers = convert_content_data_to_df(data)

    # then
    assert list(df_name_as_headers) == ["a", "b"]
    assert df_name_as_headers.size == 4
    assert (df_name_as_headers.get("a") == [1, 3]).all()
    assert (df_name_as_headers.get("b") == [2, 4]).all()

    assert list(df_title_as_headers) == ["A1", "B1"]
    assert df_title_as_headers.size == 4
    assert (df_title_as_headers.get("A1") == [1, 3]).all()
    assert (df_title_as_headers.get("B1") == [2, 4]).all()

    assert list(df_default_headers) == ["a", "b"]
    assert df_default_headers.size == 4
    assert (df_default_headers.get("a") == [1, 3]).all()
    assert (df_default_headers.get("b") == [2, 4]).all()


@pytest.mark.parametrize(
    "input_arg",
    [[1, 2, 3], tuple([1, 2, 3]), {1, 2, 3}, {1: "1", 2: "2", 3: "3"}],
)
def test_convert_to_list_iterable_objects(input_arg):
    # given
    id_input_arg = id(input_arg)

    # when
    test_result = try_copy_to_list(input_arg)

    # then
    assert test_result == list(input_arg)
    assert id(test_result) != id_input_arg


@pytest.mark.parametrize(
    "input_arg",
    [None, 1, True, 1.1, "", "sdsads"],
)
def test_convert_to_list_not_iterable_objects(input_arg):
    # given
    id_input_arg = id(input_arg)

    # when
    test_result = try_copy_to_list(input_arg)

    # then
    assert test_result == input_arg
    assert id(test_result) == id_input_arg



from ._basis_spline_smooth_model import *
from ._business_sector import *
from ._calendar_adjustment import *
from ._calibration_model import *
from ._curve_sub_type import *
from ._economic_sector import *
from ._industry import *
from ._industry_group import *
from ._interest_calculation_method import *
from ._interpolation_mode import *
from ._issuer_type import *
from ._main_constituent_asset_class import *
from ._price_side import *
from ._rating import *
from ._rating_scale_source import *
from ._reference_entity_type import *
from ._seniority import *



# coding: utf8
from typing import TYPE_CHECKING

from ._forward_curve_definition import ForwardCurveDefinition
from ._swap_zc_curve_definition import SwapZcCurveDefinition
from ._swap_zc_curve_parameters import SwapZcCurveParameters
from .._object_definition import ObjectDefinition
from ...._types import OptStr
from ...._tools import ArgsParser

if TYPE_CHECKING:
    from ._forward_curve_types import (
        CurveDefinition,
        CurveParameters,
        ForwardCurveDefinitions,
    )


def parse_forward_curve_definitions(
    param: "ForwardCurveDefinitions",
) -> "ForwardCurveDefinitions":
    if not param:
        return param

    if not isinstance(param, list):
        param = [param]

    return param


forward_curve_definitions_arg_parser = ArgsParser(parse_forward_curve_definitions)


class ForwardCurveRequestItem(ObjectDefinition):
    """
    Parameters
    ----------
    curve_definition : SwapZcCurveDefinition, optional

    curve_parameters : SwapZcCurveParameters, optional

    forward_curve_definitions : list of ForwardCurveDefinition, optional

    curve_tag : str, optional

    """

    def __init__(
        self,
        curve_definition: "CurveDefinition" = None,
        forward_curve_definitions: "ForwardCurveDefinitions" = None,
        curve_parameters: "CurveParameters" = None,
        curve_tag: OptStr = None,
    ) -> None:
        super().__init__()
        self.curve_definition = curve_definition
        self.curve_parameters = curve_parameters
        self.forward_curve_definitions = forward_curve_definitions_arg_parser.parse(
            forward_curve_definitions
        )
        self.curve_tag = curve_tag

    @property
    def curve_definition(self):
        """
        :return: object SwapZcCurveDefinition
        """
        return self._get_object_parameter(SwapZcCurveDefinition, "curveDefinition")

    @curve_definition.setter
    def curve_definition(self, value):
        self._set_object_parameter(SwapZcCurveDefinition, "curveDefinition", value)

    @property
    def curve_parameters(self):
        """
        :return: object SwapZcCurveParameters
        """
        return self._get_object_parameter(SwapZcCurveParameters, "curveParameters")

    @curve_parameters.setter
    def curve_parameters(self, value):
        self._set_object_parameter(SwapZcCurveParameters, "curveParameters", value)

    @property
    def forward_curve_definitions(self):
        """
        :return: list ForwardCurveDefinition
        """
        return self._get_list_parameter(
            ForwardCurveDefinition, "forwardCurveDefinitions"
        )

    @forward_curve_definitions.setter
    def forward_curve_definitions(self, value):
        self._set_list_parameter(
            ForwardCurveDefinition, "forwardCurveDefinitions", value
        )

    @property
    def curve_tag(self):
        """
        :return: str
        """
        return self._get_parameter("curveTag")

    @curve_tag.setter
    def curve_tag(self, value):
        self._set_parameter("curveTag", value)



from dataclasses import dataclass

import pandas as pd


@dataclass
class Subcategory:
    name: str
    revision_id: int
    revision_date: str
    top_news_id: str

    @classmethod
    def from_dict(cls, datum: dict) -> "Subcategory":
        return cls(
            name=datum.get("name"),
            revision_id=datum.get("revisionId"),
            revision_date=pd.to_datetime(datum.get("revisionDate")),
            top_news_id=datum.get("topNewsId"),
        )



from unittest.mock import patch, MagicMock

import pytest

from refinitiv.data.content.fundamental_and_reference._data_provider import (
    DataGridRDPRequestFactory,
    DataGridUDFRequestFactory,
    DataGridRDPResponseFactory,
    DataGridUDFResponseFactory,
    DataGridRDPContentValidator,
    DataGridUDFContentValidator,
    validate_correct_format_parameters,
)
from refinitiv.data.delivery._data._data_provider import ParsedData
from refinitiv.data.delivery._data._endpoint_data import Error, RequestMethod
from tests.unit.conftest import StubSession, kwargs
from .conftest import (
    ExpectedArgs,
    side_effect_func,
    PATH_MOCK_FUNC,
)


@patch(PATH_MOCK_FUNC, return_value={})
def tests_data_grid_rdp_request_factory_get_body_parameters(mock_func):
    # given
    factory = DataGridRDPRequestFactory()

    # when
    testing_value = factory.get_body_parameters()

    # then
    assert testing_value == {}


@patch(PATH_MOCK_FUNC, return_value=MagicMock(side_effect=side_effect_func))
def tests_data_grid_rdp_request_factory_get_body_parameters_pass_universe(mock_func):
    # given
    factory = DataGridRDPRequestFactory()

    # when
    testing_value = factory.get_body_parameters(universe="universe")

    # then
    assert testing_value["universe"] is not None


@patch(PATH_MOCK_FUNC, return_value=MagicMock(side_effect=side_effect_func))
def tests_data_grid_rdp_request_factory_get_body_parameters_pass_fields(mock_func):
    # given
    factory = DataGridRDPRequestFactory()

    # when
    testing_value = factory.get_body_parameters(fields="fields")

    # then
    assert testing_value["fields"] is not None


@patch(PATH_MOCK_FUNC, return_value=MagicMock(side_effect=side_effect_func))
def tests_data_grid_rdp_request_factory_get_body_parameters_pass_parameters(mock_func):
    # given
    factory = DataGridRDPRequestFactory()

    # when
    testing_value = factory.get_body_parameters(parameters="parameters")

    # then
    assert testing_value["parameters"] is not None


@patch(PATH_MOCK_FUNC)
def tests_data_grid_rdp_request_factory_get_body_parameters_pass_output(mock_func):
    # given
    output = {"layout": {"output": ...}}
    mock_func.return_value = output

    factory = DataGridRDPRequestFactory()

    # when
    testing_value = factory.get_body_parameters(output=output)

    # then
    assert testing_value["output"] is not None


@pytest.mark.parametrize(
    "input_layout",
    [
        {"layout": ...},
        {"layout": {"out": ...}},
        {"layout": {"output": None}},
        {"layout": {"output": ""}},
        {"layout": {"output": []}},
        {"layout": {"output": set()}},
        {"layout": {"output": {}}},
    ],
)
@patch(PATH_MOCK_FUNC)
def tests_data_grid_rdp_request_factory_get_body_parameters_fail_output(
    mock_func, input_layout
):
    # given
    mock_func.return_value = input_layout

    factory = DataGridRDPRequestFactory()

    # when
    testing_value = factory.get_body_parameters(layout=input_layout)

    # then
    assert "output" not in testing_value


def tests_data_grid_rdp_request_factory_get_request_method():
    # given
    factory = DataGridRDPRequestFactory()

    # when
    testing_value = factory.get_request_method()

    # then
    assert testing_value == RequestMethod.POST


@patch(PATH_MOCK_FUNC, return_value=MagicMock(side_effect=side_effect_func))
def tests_data_grid_udf_request_factory_get_body_parameters_pass_instruments(mock_func):
    # given
    factory = DataGridUDFRequestFactory()

    # when
    testing_value = factory.get_body_parameters(universe="instruments")

    # then
    assert testing_value["instruments"] is not None


@patch(PATH_MOCK_FUNC, return_value=MagicMock(side_effect=side_effect_func))
def tests_data_grid_udf_request_factory_get_body_parameters_pass_fields(mock_func):
    # given
    factory = DataGridUDFRequestFactory()

    # when
    testing_value = factory.get_body_parameters(fields="fields")

    # then
    assert testing_value["fields"] is not None


@patch(PATH_MOCK_FUNC)
def tests_data_grid_udf_request_factory_get_body_parameters_pass_layout(mock_func):
    # given
    layout = {"layout": {"layout": ...}}
    mock_func.return_value = layout

    factory = DataGridUDFRequestFactory()

    # when
    testing_value = factory.get_body_parameters(layout=layout)

    # then
    assert testing_value["layout"] is not None


@pytest.mark.parametrize(
    "input_layout",
    [
        {"layout": ...},
        {"layout": {"a": ...}},
        {"layout": {"layout": None}},
        {"layout": {"layout": ""}},
        {"layout": {"layout": []}},
        {"layout": {"layout": set()}},
        {"layout": {"layout": {}}},
    ],
)
@patch(PATH_MOCK_FUNC)
def tests_data_grid_udf_request_factory_get_body_parameters_fail_layout(
    mock_func, input_layout
):
    # given
    mock_func.return_value = input_layout

    factory = DataGridUDFRequestFactory()

    # when
    testing_value = factory.get_body_parameters(layout=input_layout)

    # then
    assert "layout" not in testing_value


@patch(PATH_MOCK_FUNC)
def tests_data_grid_udf_request_factory_get_body_parameters_skip_tr_fields(mock_func):
    # given
    factory = DataGridUDFRequestFactory()
    input_fields = ["TR.PriceClose", "TR.Volume", "TR.PriceLow", "BID", "ASK"]
    expected_fields = [
        {"name": "TR.PriceClose"},
        {"name": "TR.Volume"},
        {"name": "TR.PriceLow"},
    ]
    mock_func.return_value = {
        "fields": input_fields,
        "expected_fields": expected_fields,
    }

    # when
    testing_value = factory.get_body_parameters(fields=input_fields)

    # then
    assert testing_value["fields"] == expected_fields


@patch(PATH_MOCK_FUNC)
def tests_data_grid_udf_request_factory_get_body_parameters_add_tr_and_func_fields(
    mock_func,
):
    # given
    factory = DataGridUDFRequestFactory()
    input_fields = [
        "TR.PriceClose",
        "TR.Volume",
        "tR.PriceLow",
        "BID",
        "ASK",
        "Tr.PeersRank",
        "tr.RIC",
        "AVAIL(TR.GrossProfit(Period=LTM,Methodology=InterimSum)/TR.TotalRevenue(Period=LTM,Methodology=InterimSum),TR.GrossProfit(Period=FY0)/TR.TotalRevenue(Period=FY0))*100",
    ]
    expected_fields = [
        {"name": "TR.PriceClose"},
        {"name": "TR.Volume"},
        {"name": "tR.PriceLow"},
        {"name": "Tr.PeersRank"},
        {"name": "tr.RIC"},
        {
            "name": "AVAIL(TR.GrossProfit(Period=LTM,Methodology=InterimSum)/TR.TotalRevenue(Period=LTM,Methodology=InterimSum),TR.GrossProfit(Period=FY0)/TR.TotalRevenue(Period=FY0))*100"
        },
    ]
    mock_func.return_value = {
        "fields": input_fields,
        "expected_fields": expected_fields,
    }

    # when
    testing_value = factory.get_body_parameters(fields=input_fields)

    # then
    assert testing_value["fields"] == expected_fields


@patch(PATH_MOCK_FUNC, return_value=MagicMock(side_effect=side_effect_func))
def tests_data_grid_udf_request_factory_get_body_parameters_pass_parameters(mock_func):
    # given
    factory = DataGridUDFRequestFactory()

    # when
    testing_value = factory.get_body_parameters(parameters="parameters")

    # then
    assert testing_value["parameters"] is not None


def tests_data_grid_udf_request_factory_get_request_method():
    # given
    factory = DataGridUDFRequestFactory()

    # when
    testing_value = factory.get_request_method()

    # then
    assert testing_value == RequestMethod.POST


@patch(PATH_MOCK_FUNC, return_value=MagicMock(side_effect=side_effect_func))
def tests_data_grid_udf_request_factory_create(mock_func):
    # given
    factory = DataGridUDFRequestFactory()
    session = StubSession(is_open=True)

    # when
    data = factory.create(session)

    # then
    assert data is not None


def test_data_grid_udf_response_factory_create_success():
    # given
    data = ParsedData({}, {}, **{"content_data": {"responses": [{}]}})
    response_factory = DataGridUDFResponseFactory()

    # when
    result = response_factory.create_success(data)

    # then
    assert hasattr(result, "data")
    assert result.data.raw == data.content_data["responses"][0]


@pytest.mark.parametrize(
    "data,kwargs,error_code, error_message",
    [
        (
            ({"error_codes": 412}),
            {"universe": ["INVALID"], "fields": ["TR.Volume"]},
            412,
            "Unable to resolve all requested identifiers in ['INVALID'].",
        ),
        (
            ({"error_codes": 218}),
            {"universe": "IBM.N", "fields": ["TR.Volumeee"]},
            218,
            "Unable to resolve all requested fields in ['TR.Volumeee']. "
            "The formula must contain at least one field or function.",
        ),
        (
            ({"error_codes": 101010, "error_messages": "It's error message"}),
            {"universe": ["IBM.N"], "fields": ["TR.Volume"]},
            101010,
            "It's error message Requested universes: ['IBM.N']. "
            "Requested fields: ['TR.Volume']",
        ),
    ],
)
def test_data_grid_udf_response_factory_create_fail(
    data, kwargs, error_code, error_message
):
    parsed_data = ParsedData({}, {}, **data)
    response_factory = DataGridUDFResponseFactory()
    result = response_factory.create_fail(parsed_data, **kwargs)

    assert hasattr(result, "errors")
    assert isinstance(result.errors[0], Error)
    assert result.errors[0][0] == error_code
    assert result.errors[0][1] == error_message


def test_data_grid_udf_response_factory_create_success_with_err():
    data = ParsedData(
        {},
        {},
        **{
            "content_data": {
                "responses": [
                    {
                        "error": [
                            {
                                "code": 416,
                                "col": 2,
                                "message": "Unable to collect data for the field "
                                "'TR.Volume' and some specific identifier(s).",
                                "row": 1,
                            }
                        ]
                    }
                ]
            }
        },
    )
    expected_result = [
        (
            416,
            "Unable to collect data for the field 'TR.Volume' "
            "and some specific identifier(s).",
        )
    ]
    response_factory = DataGridUDFResponseFactory()
    result = response_factory.create_success(data)
    assert result.errors == expected_result


def test_data_grid_rdp_response_factory_create_success():
    # given
    data = ParsedData({}, {}, {"messages": {"descriptions": ""}})
    response_factory = DataGridRDPResponseFactory()

    # when
    result = response_factory.create_success(data)

    # then
    assert hasattr(result, "data")
    assert result.data.raw == data.content_data


def test_data_grid_rdp_response_factory_create_success_with_err():
    data = ParsedData(
        {},
        {},
        {
            "messages": {
                "descriptions": [
                    {
                        "code": 416,
                        "description": "Unable to collect data for the field "
                        "'TR.Volume' and some specific identifier(s).",
                    }
                ]
            }
        },
    )
    expected_result = [
        (
            416,
            "Unable to collect data for the field 'TR.Volume' "
            "and some specific identifier(s).",
        )
    ]
    response_factory = DataGridRDPResponseFactory()
    result = response_factory.create_success(data)
    assert result.errors == expected_result


@pytest.mark.parametrize(
    "data,kwargs,error_code, error_message",
    [
        (
            ({"error_codes": 412}),
            {"universe": ["INVALID"], "fields": ["TR.Volume"]},
            412,
            "Unable to resolve all requested identifiers in ['INVALID'].",
        ),
        (
            ({"error_codes": 218}),
            {"universe": "IBM.N", "fields": ["TR.Volumeee"]},
            218,
            "Unable to resolve all requested fields in ['TR.Volumeee']. "
            "The formula must contain at least one field or function.",
        ),
        (
            ({"error_codes": 101010, "error_messages": "It's error message"}),
            {"universe": ["IBM.N"], "fields": ["TR.Volume"]},
            101010,
            "It's error message Requested universes: ['IBM.N']. "
            "Requested fields: ['TR.Volume']",
        ),
    ],
)
def test_data_grid_rdp_response_factory_create_fail(
    data, kwargs, error_code, error_message
):
    parsed_data = ParsedData({}, {}, **data)
    response_factory = DataGridRDPResponseFactory()
    result = response_factory.create_fail(parsed_data, **kwargs)

    assert hasattr(result, "errors")
    assert isinstance(result.errors[0], Error)
    assert result.errors[0][0] == error_code
    assert result.errors[0][1] == error_message


@pytest.mark.parametrize(
    "input_kwargs, expected_values",
    [
        (
            kwargs(universe="", fields=""),
            ExpectedArgs(universe=[""], fields=[""]),
        ),
        (
            kwargs(universe="DD", fields=""),
            ExpectedArgs(universe=["DD"], fields=[""]),
        ),
        (
            kwargs(universe="dd", fields=""),
            ExpectedArgs(universe=["DD"], fields=[""]),
        ),
        (
            kwargs(universe=["dd"], fields=""),
            ExpectedArgs(universe=["DD"], fields=[""]),
        ),
        (
            kwargs(universe=["dd"], fields="ss"),
            ExpectedArgs(universe=["DD"], fields=["ss"]),
        ),
        (
            kwargs(universe=["dd"], fields=["ss"]),
            ExpectedArgs(universe=["DD"], fields=["ss"]),
        ),
    ],
)
def test_validate_correct_format_parameters_args_universe_and_fields(
    input_kwargs, expected_values
):
    # when
    result = validate_correct_format_parameters(
        use_field_names_in_headers=False, **input_kwargs
    )

    # then
    assert result["universe"] == expected_values.universe
    assert result["fields"] == expected_values.fields


@pytest.mark.parametrize(
    "input_kwargs, expected_raise",
    [
        (kwargs(universe="", fields=""), ValueError),
        (kwargs(universe=1, fields=""), TypeError),
        (kwargs(universe={}, fields=""), TypeError),
        (kwargs(universe=[1, 2, 3], fields=""), ValueError),
        (kwargs(universe="", fields=1), TypeError),
        (kwargs(universe="", fields={}), TypeError),
        (kwargs(universe="", fields=[1, 2, 3]), ValueError),
        (kwargs(universe=["1", "2"], fields=["1", "2"]), ValueError),
        (kwargs(universe=["1"], fields=["1"], parameters=""), ValueError),
        (kwargs(universe=["1"], fields=["1"], parameters="d"), ValueError),
        (kwargs(universe=["1"], fields=["1"], parameters=[]), ValueError),
        (kwargs(universe=["1"], fields=["1"], parameters=[1]), ValueError),
        (kwargs(universe=["1"], fields=["1"], extended_params="1"), AttributeError),
        (kwargs(universe=["1"], fields=["1"], extended_params=[0]), AttributeError),
        (kwargs(universe=["1"], fields=["1"], extended_params=True), AttributeError),
        (
            kwargs(universe=["1"], fields=["1"], use_field_names_in_headers=""),
            ValueError,
        ),
        (
            kwargs(universe=["1"], fields=["1"], use_field_names_in_headers=" "),
            ValueError,
        ),
        (
            kwargs(universe=["1"], fields=["1"], use_field_names_in_headers=0),
            ValueError,
        ),
        (
            kwargs(universe=["1"], fields=["1"], use_field_names_in_headers=None),
            ValueError,
        ),
    ],
)
def test_validate_correct_format_parameters_raise_exception(
    input_kwargs, expected_raise
):
    # when
    with pytest.raises(expected_raise):
        validate_correct_format_parameters(**input_kwargs)


@pytest.mark.parametrize(
    "input_value, expected_value",
    [
        (None, {}),
        ("", {}),
        ([], {}),
        ({}, {}),
        ({"universe": ["1", "2"]}, {"universe": ["1", "2"]}),
    ],
)
def test_validate_correct_format_parameters_arg_extended_params(
    input_value, expected_value
):
    # when
    result = validate_correct_format_parameters(
        universe=["1"],
        fields=["1"],
        use_field_names_in_headers=False,
        extended_params=input_value,
    )

    # then
    assert result["extended_params"] == expected_value


@pytest.mark.parametrize(
    "input_value, expected_universe, expected_extended_params",
    [
        (None, ["1"], {}),
        ("", ["1"], {}),
        ({}, ["1"], {}),
        ({"universe": None}, ["1"], {"universe": None}),
        ({"universe": ""}, ["1"], {"universe": ""}),
        ({"universe": []}, ["1"], {"universe": []}),
        ({"universe": ["1", "2"]}, ["1", "2"], {"universe": ["1", "2"]}),
    ],
)
def test_validate_correct_format_parameters_swap_universe(
    input_value, expected_universe, expected_extended_params
):
    # when
    result = validate_correct_format_parameters(
        universe=["1"],
        fields=["1"],
        use_field_names_in_headers=False,
        extended_params=input_value,
    )

    # then
    assert result["universe"] == expected_universe
    assert result["extended_params"] == expected_extended_params


@pytest.mark.parametrize(
    "input_content_data",
    [
        None,
        {"error": {"a": "a"}, "data": []},
        {"error": {"description": "description"}, "data": []},
        {"error": {"code": 1, "description": "description"}, "data": []},
    ],
)
def test_data_grid_rdp_content_validator_validate_content_data_is_false(
    input_content_data,
):
    # given
    input_value = ParsedData({}, {}, **{"content_data": input_content_data})
    expected_value = False
    validator = DataGridRDPContentValidator()

    # when
    testing_value = validator.validate(input_value)

    # then
    assert testing_value == expected_value


def test_data_grid_rdp_content_validator_validate_content_is_false():
    # given
    input_value = ParsedData({"content": "Failed ..."}, {})
    expected_value = False
    validator = DataGridRDPContentValidator()

    # when
    testing_value = validator.validate(input_value)

    # then
    assert testing_value == expected_value


@pytest.mark.parametrize(
    "input_content_data",
    [
        {"status": {"code": "Error"}},
        {"status": {"code": "UserRequestError"}},
        {"error": {"code": 1, "description": "description"}, "data": [1, 2, 3]},
    ],
)
def test_data_grid_rdp_content_validator_validate_content_data_is_true(
    input_content_data,
):
    # given
    input_value = ParsedData({}, {}, input_content_data)
    expected_value = True
    validator = DataGridRDPContentValidator()

    # when
    testing_value = validator.validate(input_value)

    # then
    assert testing_value == expected_value


@pytest.mark.parametrize(
    "input_content_data",
    [
        {},
        {"ErrorCode": 1, "ErrorMessage": "ErrorMessage"},
        {"responses": [{"error": "error"}]},
        {"responses": [{"error": "error", "data": ""}]},
        {"responses": [{"error": "error", "data": []}]},
        {"responses": [{"error": {"code": "code"}}]},
        {"responses": [{"error": {"code": "code", "message": "message"}}]},
        {"responses": [{"error": {"code": "code", "message": "message"}, "data": ""}]},
        {"responses": [{"error": {"code": "code", "message": "message"}, "data": []}]},
    ],
)
def test_data_grid_udf_content_validator_validate_content_data_is_false(
    input_content_data,
):
    # given
    input_value = {"status": {}, "content_data": input_content_data}
    expected_value = False
    validator = DataGridUDFContentValidator()

    # when
    testing_value = validator.validate(input_value)

    # then
    assert testing_value == expected_value


def test_data_grid_udf_content_validator_validate_content_is_false():
    # given
    input_value = ParsedData({"content": "Failed ..."}, {})
    expected_value = False
    validator = DataGridUDFContentValidator()

    # when
    testing_value = validator.validate(input_value)

    # then
    assert testing_value == expected_value


@pytest.mark.parametrize(
    "input_content_data",
    [
        {"responses": [{"data": [1]}]},
    ],
)
def test_data_grid_udf_content_validator_validate_content_data_is_true(
    input_content_data,
):
    # given
    input_value = ParsedData({}, {}, input_content_data)
    expected_value = True
    validator = DataGridUDFContentValidator()

    # when
    testing_value = validator.validate(input_value)

    # then
    assert testing_value == expected_value


@pytest.mark.parametrize(
    "input_value",
    [
        {"status": {}, "raw_response": {}},
        {"status": {"content": "Failed"}, "raw_response": {}},
        {"status": {}, "raw_response": {}, "content_data": "ErrorCode"},
        {"status": {}, "raw_response": {}, "content_data": {"ErrorCode": ...}},
        {
            "status": {},
            "raw_response": {},
            "content_data": {
                "responses": [{"error": {"code": "code", "message": "message"}}]
            },
        },
        {
            "status": {},
            "raw_response": {},
            "content_data": {
                "responses": [
                    {"error": {"code": "code", "message": "message"}, "data": None}
                ]
            },
        },
        {
            "status": {},
            "raw_response": {},
            "content_data": {
                "responses": [
                    {"error": {"code": "code", "message": "message"}, "data": ""}
                ]
            },
        },
        {
            "status": {},
            "raw_response": {},
            "content_data": {
                "responses": [
                    {"error": {"code": "code", "message": "message"}, "data": []}
                ]
            },
        },
        {
            "status": {},
            "raw_response": {},
            "content_data": {
                "responses": [
                    {
                        "error": [{"code": "code", "message": "message"}],
                        "data": [1],
                        "headers": [[{"displayName": None}]],
                    }
                ]
            },
        },
    ],
)
def test_data_grid_udf_content_validator_validate_content_data_is_false(input_value):
    # given
    expected_value = False
    validator = DataGridUDFContentValidator()

    # when
    parsed_data = ParsedData(**input_value)
    testing_value = validator.validate(parsed_data)

    # then
    assert testing_value == expected_value


def test_data_grid_udf_content_validator_validate_content_data_is_str():
    # given
    input_value = ParsedData({}, {}, **{"content_data": "some text html"})
    expected_value = False
    validator = DataGridUDFContentValidator()

    # when
    testing_value = validator.validate(input_value)

    # then
    assert testing_value == expected_value
    assert input_value.first_error_message == input_value.content_data



import abc
import dataclasses
from copy import deepcopy
from functools import partial
from itertools import product
from typing import List, Any, Dict, Tuple

import pandas as pd
from pandas.core.tools.datetimes import DatetimeScalarOrArrayConvertible

from .._tools import ADC_FUNC_PATTERN_IN_FIELDS
from .._tools._dataframe import convert_str_to_datetime, convert_dtypes


@dataclasses.dataclass
class CacheItem:
    """Data cache item."""

    fields_by_inst: Dict[str, List[Any]] = dataclasses.field(
        init=False, default_factory=dict
    )

    def add(self, inst: str, fields: List[Any]):
        """
        Add cache item as an instance:fields key-value pair.

        Parameters
        ----------
        inst : str
            Instance name.
        fields : List[Any]
            Instance fields.
        """
        self.fields_by_inst[inst] = fields

    def has(self, inst: str) -> bool:
        """
        Boolean that indicates is the instance stored in cache item or not.

        Parameters
        ----------
        inst : str
            Instance name.

        Returns
        -------
        bool
            True if instance is already stored, False otherwise.

        """
        return inst in self.fields_by_inst


@dataclasses.dataclass
class CacheItems:
    """Container for CacheItem instances."""

    _items: List[CacheItem] = dataclasses.field(init=False, default_factory=list)

    def add(self, inst: str, fields: List[Any]):
        """Add item as a key-value pair to CacheItems container.

        Parameters
        ----------
        inst : str
            Instance name.
        fields : List[Any]
            Instance fields.
        """
        item = CacheItem()
        item.add(inst, fields)
        self._items.append(item)

    def get_item_without(self, inst: str):
        """
        Gets cache item without particular instance.

        Parameters
        ----------
        inst : str
            Instance name.

        Returns
        -------
        item : CacheItem
            CacheItem that does not contain particular instance.

        """
        for item in self._items:
            if not item.has(inst):
                return item

        raise ValueError(f"Cannot get item without inst={inst}.")

    def __iter__(self):
        self._n = 0
        return self

    def __next__(self):
        if self._n < len(self._items):
            result = self._items[self._n]
            self._n += 1
            return result
        raise StopIteration


@dataclasses.dataclass
class DateCache:
    """
    Cache for data items that belong to particular date.

    DateCache has the following structure:

    Example
    -------
    DateCache {
        'date_1': CacheItems [
            CacheItem {
                dict {
                    'inst_A': [1, 2, 3],
                    'inst_B': [1, 2, 3]
                }
            },
            CacheItem {
                dict {
                    'inst_A': [1, 2, 3]
                }
            }
        ]
    }
    """

    _cache: Dict[str, CacheItems] = dataclasses.field(init=False, default_factory=dict)

    def can_update_fields(self, date: str, inst: str) -> bool:
        """
        Boolean that indicates can item fields be updated or not.

        Parameters
        ----------
        date : str
            Date string to retrieve a bunch of items that belong to this date.
        inst : str
            Instance name to check if CacheItems has this instance or not.

        Returns
        -------
        bool
            True if fields can be updated, False otherwise
        """
        for item in self._cache.get(date, CacheItems()):
            if not item.has(inst):
                return True

        return False

    def add(self, date: str, inst: str, fields: List[Any]):
        """Add items to CacheItems container in data cache during initialization.

        Parameters
        ----------
        date : str
            Date string.
        inst : str
            Instance name.
        fields : List[Any]
            Instance fields, filled according template.
        """
        items = self._cache.setdefault(date, CacheItems())
        items.add(inst, fields)

    def update_fields(
        self,
        date: str,
        inst: str,
        fields: List[Any],
        num_columns: int,
        unique_insts: List[str],
    ) -> List[Any]:
        """Updates fields in data cache and returns updated ones.

        Parameters
        ----------
        date : str
            Date string.
        inst : str
            Instance name.
        fields : List[Any]
            Fields list.
        num_columns : int
            Data columns quantity.
        unique_insts : List[str]
            List of unique instances across dataframe.

        Returns
        -------
        cache_fields : List[Any]
            List of updated fields.
        """
        fields_by_inst = self._cache[date].get_item_without(inst).fields_by_inst
        idx = max(unique_insts.index(inst) for inst in fields_by_inst.keys())
        cache_inst = unique_insts[idx]
        cache_fields = fields_by_inst[cache_inst]
        idx = unique_insts.index(inst)
        cache_idx = unique_insts.index(cache_inst)

        left_idx = cache_idx * num_columns + num_columns
        right_idx = idx * num_columns + num_columns

        for idx in range(left_idx, right_idx):
            cache_fields[idx] = fields[idx]

        fields_by_inst[inst] = cache_fields
        return cache_fields


def partial_process_index(
    num_unique_insts: int,
    num_columns: int,
    date_cache: DateCache,
    index: List[DatetimeScalarOrArrayConvertible],
    unique_insts: List[str],
    inst: str,
    date: str,
    fields: List[Any],
):
    is_add = True

    if num_unique_insts > 1:
        total = num_unique_insts * num_columns
        template = [pd.NA] * total

        idx = unique_insts.index(inst)
        right_idx = idx * num_columns + num_columns
        left_idx = idx * num_columns
        for item, idx in zip(fields, range(left_idx, right_idx)):
            template[idx] = item
        fields = template

        if date_cache.can_update_fields(date, inst):
            is_add = False
            fields = date_cache.update_fields(
                date, inst, fields, num_columns, unique_insts
            )

        else:
            date_cache.add(date, inst, fields)
            index.append(convert_str_to_datetime(date))

    else:
        index.append(convert_str_to_datetime(date))

    return fields, is_add


class DFBuilder(abc.ABC):
    @staticmethod
    def get_header_key(use_field_names_in_headers: bool) -> str:
        return "name" if use_field_names_in_headers else "title"

    @classmethod
    def is_date_column(cls, _: dict, column: str) -> bool:
        return "date" in column.lower()

    @abc.abstractmethod
    def get_instrument_column_name(self, header_key: str) -> str:
        # for override
        pass

    @abc.abstractmethod
    def get_date_column_name(self, header_key: str) -> str:
        # for override
        pass

    @abc.abstractmethod
    def get_headers(self, content_data: dict) -> List[dict]:
        # for override
        pass

    def get_date_idxs_and_columns(
        self, headers: List[dict], use_field_names_in_headers: bool = False
    ) -> Tuple[List[str], List[int]]:
        header_key = self.get_header_key(use_field_names_in_headers)
        columns = []
        date_idxs = []
        for idx, header in enumerate(headers):
            col = header[header_key]
            columns.append(col)
            if self.is_date_column(header, col):
                date_idxs.append(idx)

        return columns, date_idxs

    def build_index(
        self, content_data: dict, use_field_names_in_headers: bool = False, **kwargs
    ) -> pd.DataFrame:
        data = []
        columns, date_idxs = self.get_date_idxs_and_columns(
            self.get_headers(content_data), use_field_names_in_headers
        )
        for fields in content_data.get("data", []):
            fields = list(fields)

            for idx, item in enumerate(fields):
                if item is None:
                    fields[idx] = pd.NA

            for idx in date_idxs:
                fields[idx] = convert_str_to_datetime(fields[idx])

            data.append(fields)

        df = pd.DataFrame(data=data, columns=columns)
        df = convert_dtypes(df)
        return df

    def build_date_as_index(
        self,
        content_data: dict,
        use_field_names_in_headers: bool = False,
        use_multiindex: bool = False,
        **kwargs,
    ) -> pd.DataFrame:
        header_key = self.get_header_key(use_field_names_in_headers)
        instrument_column_name = self.get_instrument_column_name(header_key)
        date_column_name = self.get_date_column_name(header_key)
        columns = []
        date_idxs = []
        inst_idx = None
        date_idx = None
        headers = self.get_headers(content_data)
        for idx, header in enumerate(headers):
            col = header[header_key]
            if inst_idx is None and col == instrument_column_name:
                inst_idx = idx
                continue

            if date_idx is None and col == date_column_name:
                date_idx = idx
                continue

            columns.append(col)
            if self.is_date_column(header, col):
                date_idxs.append(columns.index(col))

        data = []
        index = []
        num_columns = len(columns)
        date_cache = DateCache()
        fields_list = content_data.get("data", [])
        unique_insts = list(dict.fromkeys(fields[inst_idx] for fields in fields_list))
        num_unique_insts = len(unique_insts)
        process_index = partial(
            partial_process_index,
            num_unique_insts,
            num_columns,
            date_cache,
            index,
            unique_insts,
        )
        for fields in fields_list:
            fields = list(fields)
            date_str = fields[date_idx]

            if not date_str:
                continue

            inst = fields[inst_idx]
            fields.pop(date_idx)
            fields.pop(inst_idx)

            fields = [pd.NA if i is None else i for i in fields]

            for idx in date_idxs:
                fields[idx] = convert_str_to_datetime(fields[idx])

            fields, is_add = process_index(inst, date_str, fields)

            is_add and data.append(fields)

        if num_columns > 1 and num_unique_insts > 1 or use_multiindex:
            columns = pd.MultiIndex.from_tuples(product(unique_insts, columns))

        elif num_unique_insts == 1:
            columns = pd.Index(data=columns, name=unique_insts.pop())

        elif num_columns == 1:
            columns = pd.Index(data=unique_insts, name=columns.pop())

        index = pd.Index(data=index, name=date_column_name)
        df = pd.DataFrame(data=data, columns=columns, index=index)
        df = convert_dtypes(df)
        df.sort_index(ascending=False, inplace=True)
        return df


class DFBuilderRDP(DFBuilder):
    """
    {
        "links": {"count": 2},
        "variability": "",
        "universe": [
            {
                "Instrument": "GOOG.O",
                "Company Common Name": "Alphabet Inc",
                "Organization PermID": "5030853586",
                "Reporting Currency": "USD",
            }
        ],
        "data": [
            ["GOOG.O", "2022-01-26T00:00:00", "USD", None],
            ["GOOG.O", "2020-12-31T00:00:00", None, "2020-12-31T00:00:00"],
        ],
        "messages": {
            "codes": [[-1, -1, -1, -2], [-1, -1, -2, -1]],
            "descriptions": [
                {"code": -2, "description": "empty"},
                {"code": -1, "description": "ok"},
            ],
        },
        "headers": [
            {
                "name": "instrument",
                "title": "Instrument",
                "type": "string",
                "description": "The requested Instrument as defined by the user.",
            },
            {
                "name": "date",
                "title": "Date",
                "type": "datetime",
                "description": "Date associated with the returned data.",
            },
            {
                "name": "TR.RevenueMean",
                "title": "Currency",
                "type": "string",
                "description": "The statistical average of all broker ...",
            },
            {
                "name": "TR.Revenue",
                "title": "Date",
                "type": "datetime",
                "description": "Is used for industrial and utility companies. ...",
            },
        ],
    }
    """

    def get_instrument_column_name(self, header_key: str) -> str:
        if header_key == "name":
            return "instrument"
        elif header_key == "title":
            return "Instrument"

    def get_date_column_name(self, header_key: str) -> str:
        if header_key == "name":
            return "date"
        elif header_key == "title":
            return "Date"

    def get_headers(self, content_data) -> List[dict]:
        return content_data.get("headers", [])

    @classmethod
    def is_date_column(cls, header: dict, column: str) -> bool:
        header_type = header.get("type")
        if header_type:
            return header_type in {"datetime", "date"}
        return super().is_date_column(header, column)


class DFBuilderUDF(DFBuilder):
    """
    {
        "columnHeadersCount": 1,
        "data": [
            ["GOOG.O", "2022-01-26T00:00:00Z", "USD", ""],
            ["GOOG.O", "2020-12-31T00:00:00Z", "", "2020-12-31T00:00:00Z"],
        ],
        "headerOrientation": "horizontal",
        "headers": [
            [
                {"displayName": "Instrument"},
                {"displayName": "Date"},
                {"displayName": "Currency", "field": "TR.REVENUEMEAN.currency"},
                {"displayName": "Date", "field": "TR.REVENUE.DATE"},
            ]
        ],
        "rowHeadersCount": 2,
        "totalColumnsCount": 4,
        "totalRowsCount": 3,
    }
    """

    def get_instrument_column_name(self, column_name_key: str) -> str:
        return "Instrument"

    def get_date_column_name(self, header_key: str) -> str:
        return "Date"

    def get_headers(self, content_data) -> List[dict]:
        headers = content_data["headers"]
        headers = headers[0]
        return [
            {
                "name": header.get("field") or header.get("displayName"),
                "title": header.get("displayName"),
            }
            for header in headers
        ]

    @classmethod
    def is_date_column(cls, _: dict, column: str) -> bool:
        return "date" in column.lower() and not ADC_FUNC_PATTERN_IN_FIELDS.match(column)


def build_dates_calendars_df(raw: Any, **kwargs):
    raw = deepcopy(raw)
    add_periods_data = []

    clean_request_items = []
    for item in raw:
        if not item.get("error"):
            clean_request_items.append(item)

    for request_item in clean_request_items:
        if request_item.get("date"):
            request_item["date"] = convert_str_to_datetime(request_item["date"])

        request_item.pop("holidays", None)
        add_periods_data.append(request_item)

    _df = pd.DataFrame(add_periods_data)

    return _df


def build_dates_calendars_holidays_df(raw: Any, **kwargs):
    holidays_data = _dates_calendars_prepare_holidays_data(raw)
    holidays_df = pd.DataFrame(holidays_data)
    return holidays_df


def _dates_calendars_prepare_holidays_data(raw):
    raw = deepcopy(raw)
    holidays_data = []

    for request_item_holiday in raw:
        for holiday in request_item_holiday.get("holidays", []):
            if holiday.get("names"):
                for holiday_name in holiday["names"]:
                    holiday_name["tag"] = request_item_holiday.get("tag")
                    holiday_name["date"] = convert_str_to_datetime(holiday.get("date"))
                    holidays_data.append(holiday_name)
            else:
                holiday["tag"] = request_item_holiday.get("tag")
                holidays_data.append(holiday)
    return holidays_data


def default_build_df(raw: Any, **kwargs) -> pd.DataFrame:
    df = pd.DataFrame(raw)
    df = convert_dtypes(df)
    return df


def build_dates_calendars_date_schedule_df(raw: Any, **kwargs) -> pd.DataFrame:
    raw = deepcopy(raw)

    _dates = []
    for date in raw.get("dates"):
        _dates.append(convert_str_to_datetime(date))

    raw["dates"] = _dates
    df = pd.DataFrame(raw)
    return df


def build_empty_df(*args, **kwargs) -> pd.DataFrame:
    return pd.DataFrame()


dfbuilder_rdp = DFBuilderRDP()
dfbuilder_udf = DFBuilderUDF()



import allure
import pytest
import json

import refinitiv.data.content.ipa.financial_contracts as rdf
from refinitiv.data._content_type import ContentType
from refinitiv.data.delivery._stream import stream_cxn_cache
from refinitiv.data.errors import RDError
from tests.integration.conftest import is_open
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ipa.contracts.cap_floor.conftest import (
    cap_floor_on_cms_definition,
    cap_floor_coller_pos_definition,
    cap_floor_amortized_definition,
    invalid_cap_floor_definition,
    cap_floor_universe,
)
from tests.integration.content.ipa.contracts.conftest import (
    check_http_status_is_success_and_df_value_not_empty,
    check_stream_state_and_df_from_stream,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_response_status,
    check_dataframe_column_date_for_datetime_type,
)


@allure.suite("Content object - CapFloor")
@allure.feature("Content object - CapFloor")
@allure.severity(allure.severity_level.CRITICAL)
class TestCapFloor:
    @allure.title("Create a cap_floor definition object with params and get data")
    @pytest.mark.caseid("36325567")
    @pytest.mark.parametrize(
        "cap_floor_definition",
        [
            cap_floor_on_cms_definition,
            cap_floor_coller_pos_definition,
            cap_floor_amortized_definition,
        ],
    )
    def test_get_cap_floor_analytics(self, open_session, cap_floor_definition):
        response = cap_floor_definition().get_data()
        check_http_status_is_success_and_df_value_not_empty(response)

    @allure.title("Create a cap_floor definition without any parameters")
    @pytest.mark.caseid("36325572")
    def test_get_cap_floor_analytics_without_any_parameters(self, open_session):
        with pytest.raises(RDError) as error:
            rdf.cap_floor.Definition().get_data()
        assert (
            str(error.value)
            == "Error code QPS-DPS.1005 | Mandatory input missing: Notional currency is mandatory."
        )

    @allure.title("Create a cap_floor definition with empty pricing parameters")
    @pytest.mark.caseid("36325576")
    def test_get_cap_floor_with_empty_pricing_params(self, open_session):
        response = rdf.cap_floor.Definition(
            notional_ccy="EUR",
            start_date="2019-02-11",
            tenor="5Y",
            index_tenor="1M",
            buy_sell="Sell",
            floor_strike_percent=1,
            pricing_parameters=rdf.cap_floor.PricingParameters(),
        ).get_data()

        check_http_status_is_success_and_df_value_not_empty(response)
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title("Create a cap_floor definition with extended parameters")
    @pytest.mark.caseid("36325578")
    def test_get_cap_floor_with_extended_params(self, open_session):
        response = rdf.cap_floor.Definition(
            instrument_tag="CapGBP",
            amortization_schedule=[
                rdf.cap_floor.AmortizationItem(
                    start_date="2023-06-12",
                    end_date="2024-06-12",
                    amount=100000,
                    amortization_type="Schedule",
                    amortization_frequency=rdf.cap_floor.AmortizationFrequency.EVERY_COUPON,
                ),
                rdf.cap_floor.AmortizationItem(
                    start_date="2024-06-11",
                    end_date="2025-06-11",
                    amount=-100000,
                    amortization_type="Schedule",
                    amortization_frequency=rdf.cap_floor.AmortizationFrequency.EVERY_COUPON,
                ),
            ],
            fields=[
                "InstrumentTag",
                "InstrumentDescription",
                "FixedRate",
                "MarketDataDate",
                "MarketValueInDealCcy",
                "MarketValueInReportCcy",
                "ErrorMessage",
                "MarketDataDate",
            ],
            extended_params={
                "instrumentDefinition": {
                    "startDate": "2021-06-11",
                    "tenor": "10Y",
                    "notionalCcy": "GBP",
                    "notionalAmount": 10000000,
                    "interestPaymentFrequency": "Quarterly",
                    "buySell": "Buy",
                    "capStrikePercent": 0.25,
                },
                "pricingParameters": {
                    "marketDataDate": "2020-10-07",
                    "skipFirstCapFloorlet": False,
                    "valuationDate": "2021-10-07",
                },
            },
        ).get_data()

        check_http_status_is_success_and_df_value_not_empty(response)

        request = json.loads(response.request_message.content.decode("utf-8"))
        request_universe = request.get("universe")[0]
        assert (
            request_universe == cap_floor_universe
        ), f"Extended params are applied incorrectly"

    @allure.title("Create a cap_floor definition with invalid extended parameters")
    @pytest.mark.caseid("36325581")
    def test_get_cap_floor_with_invalid_extended_params(self, open_session):
        with pytest.raises(RDError) as error:
            rdf.cap_floor.Definition(
                notional_ccy="EUR",
                cap_strike_percent=1,
                fields=["InstrumentTag"],
                extended_params={"fields": "INVALID"},
            ).get_data()
        assert str(error.value)

    @allure.title("Create a cap_floor definition request asynchronously")
    @pytest.mark.caseid("36325638")
    async def test_cap_floor_definition_asynchronously(self, open_session_async):
        valid_response, invalid_response = await get_async_response_from_definitions(
            cap_floor_coller_pos_definition(), invalid_cap_floor_definition()
        )

        check_http_status_is_success_and_df_value_not_empty(valid_response)
        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_message="Market data error : Invalid fixing source for provided index currency. No matching fixing for INVAL",
        )

    @allure.title("Create a cap_floor stream")
    @pytest.mark.caseid("36325639")
    @pytest.mark.smoke
    def test_cap_floor_stream(self, open_session):
        stream = cap_floor_coller_pos_definition().get_stream()
        stream.open()

        check_stream_state_and_df_from_stream(stream)

    @allure.title(
        "Receiving cap_floor data when the session was interrupted during the process"
    )
    @pytest.mark.caseid("36325652")
    def test_cap_floor_data_when_session_was_interrupted(self, open_session):
        session = open_session
        stream = cap_floor_amortized_definition().get_stream()
        stream.open()
        session.close()

        is_cxn_alive = stream_cxn_cache.is_cxn_alive(
            session, ContentType.STREAMING_CONTRACTS
        )

        assert not is_cxn_alive, f"Connection is alive"
        assert not is_open(stream), f"Stream is open"

    @allure.title("Create cap_floor definition and get data for the invalid params")
    @pytest.mark.caseid("36325581")
    def test_get_cap_floor_with_invalid_params(self, open_session):
        with pytest.raises(RDError) as error:
            rdf.cap_floor.Definition(
                notional_ccy="INVALIDCCY",
                buy_sell=rdf.cap_floor.BuySell.SELL,
                cap_strike_percent=1,
                tenor="5Y",
            ).get_data()
        assert (
            str(error.value)
            == "Error code QPS-DPS.3001 | Market data error : Invalid fixing source for provided index currency. No matching fixing for INVALIDCCY"
        )



# coding: utf8

__all__ = ["BasketItem"]

from .._object_definition import ObjectDefinition


class BasketItem(ObjectDefinition):
    def __init__(self, instrument_code=None, currency=None):
        super().__init__()
        self.instrument_code = instrument_code
        self.currency = currency

    @property
    def currency(self):
        """
        :return: str
        """
        return self._get_parameter("currency")

    @currency.setter
    def currency(self, value):
        self._set_parameter("currency", value)

    @property
    def instrument_code(self):
        """
        Data Type: string
        Code to define the Reverse Convertible instrument.It can be an ISIN, a RIC or an AssetId.
        :return: str
        """
        return self._get_parameter("instrumentCode")

    @instrument_code.setter
    def instrument_code(self, value):
        self._set_parameter("instrumentCode", value)



from time import sleep

from refinitiv.data import OpenState


def check_open_and_close_stream_state(stream, wait_before_close=None):
    stream.open()
    assert stream.open_state == OpenState.Opened, "Stream is not in opened state"
    if wait_before_close is not None:
        sleep(wait_before_close)
    stream.close()
    assert stream.open_state == OpenState.Closed, "Stream is not in closed state"



# coding: utf8

__all__ = ["InterpolationWeight"]

from typing import Optional, Iterable

from ...._tools import try_copy_to_list
from .._object_definition import ObjectDefinition
from .._models import DayWeight


class InterpolationWeight(ObjectDefinition):
    """
    Parameters
    ----------
    days_list : list of DayWeight, optional

    holidays : float, optional

    week_days : float, optional

    week_ends : float, optional

    """

    def __init__(
        self,
        days_list: Optional[Iterable[DayWeight]] = None,
        holidays: Optional[float] = None,
        week_days: Optional[float] = None,
        week_ends: Optional[float] = None,
    ) -> None:
        super().__init__()
        self.days_list = try_copy_to_list(days_list)
        self.holidays = holidays
        self.week_days = week_days
        self.week_ends = week_ends

    @property
    def days_list(self):
        """
        :return: list DayWeight
        """
        return self._get_list_parameter(DayWeight, "daysList")

    @days_list.setter
    def days_list(self, value):
        self._set_list_parameter(DayWeight, "daysList", value)

    @property
    def holidays(self):
        """
        :return: float
        """
        return self._get_parameter("holidays")

    @holidays.setter
    def holidays(self, value):
        self._set_parameter("holidays", value)

    @property
    def week_days(self):
        """
        :return: float
        """
        return self._get_parameter("weekDays")

    @week_days.setter
    def week_days(self, value):
        self._set_parameter("weekDays", value)

    @property
    def week_ends(self):
        """
        :return: float
        """
        return self._get_parameter("weekEnds")

    @week_ends.setter
    def week_ends(self, value):
        self._set_parameter("weekEnds", value)



import os

import allure
import pytest

import refinitiv.data as rd
import tests.integration.conftest as conf
from refinitiv.data._core.session import set_default
from refinitiv.data.session import platform
from tests.integration.core.session.conftest import assert_session_open_state

deployed_platform_host = os.getenv("DEPLOYED_PLATFORM_HOST")
deployed_platform_user_name = os.getenv("DEPLOYED_PLATFORM_USER_NAME")


@allure.suite("Session layer - Deployed session")
@allure.feature("Session layer - Deployed session")
@allure.severity(allure.severity_level.CRITICAL)
class TestDeployedSession:
    @allure.title(
        "Verify that Deployed Platform session opens with valid credentials and valid host"
    )
    @pytest.mark.caseid("C37691450")
    @pytest.mark.smoke
    def test_session_is_opened_and_is_closed_using_valid_credentials(self):
        session = platform.Definition(
            app_key=conf.desktop_app_key,
            deployed_platform_host=deployed_platform_host,
            deployed_platform_username=deployed_platform_user_name,
            dacs_application_id="256",
        ).get_session()
        set_default(session)
        session.open()

        assert_session_open_state(session)

    @allure.title(
        "Verify that deployed session is opened using config file and is closed after close"
    )
    @pytest.mark.caseid("C43104144")
    def test_session_opened_using_config_file_and_valid_credentials(
        self, load_config
    ):
        session = platform.Definition(name="deployed-cipsnylab2").get_session()
        set_default(session)
        session.open()

        assert_session_open_state(session)

    @allure.title(
        "Verify that deployed session is opened using config file on access layer"
    )
    @pytest.mark.caseid("C43104147")
    def test_session_opened_using_config_file_in_rd_open_session(
        self, load_config
    ):
        session = rd.open_session(name="platform.deployed-cipsnylab2")

        assert_session_open_state(session)



import pandas as pd
from numpy import datetime64

from ...._tools import get_from_path
from ...._tools._dataframe import convert_dtypes

headline_data_key_by_column = {
    "headline": "text",
    "snippet": "snippet",
    "storyId": "storyId",
    "imageId": "image.id",
}


def news_top_build_df(raw: dict, **kwargs):
    raw_data = raw.get("data", [{}])

    # data
    columns = headline_data_key_by_column.keys()
    data = [
        [
            get_from_path(headline_data, path)
            for path in headline_data_key_by_column.values()
        ]
        for headline_data in raw_data
    ]

    # index
    index_data = [
        datetime64(headline_data.get("versionCreated")) for headline_data in raw_data
    ]
    index = pd.Index(index_data, name="versionCreated")
    df = pd.DataFrame(data, columns=columns, index=index)
    return convert_dtypes(df)



"""Templates functionality, specific for discovery search"""

import pandas as pd
from humps import pascalize

from refinitiv.data._tools import inspect_parameters_without_self
from refinitiv.data.content import search
from .base import Target, TargetTemplate


class DefinitionGetDataTarget(Target):
    """Template Target class that can be used with any definition to create templates"""

    definition = None

    def __init__(self):
        super().__init__()
        self.args_names = set(inspect_parameters_without_self(self.definition))

    def __call__(self, **kwargs):
        return self.definition(**kwargs).get_data().data.df


class DiscoverySearchTarget(DefinitionGetDataTarget):
    """Template Target class specific for discovery search"""

    definition = search.Definition


class DiscoverySearchTemplate(TargetTemplate):
    """Discovery search preset class"""

    _target_class = DiscoverySearchTarget

    def __repr__(self):
        return f"<DiscoverySearchTemplate '{self.name}'>"

    # redefining search only to add return type annotation
    def search(self, **kwargs) -> pd.DataFrame:
        """Please, use help() on a template object itself to get method documentation"""
        # ^ we need this docstring because we can't easily generate docstring for
        # the method, but can change __doc__ for the class instance
        return super().search(**kwargs)

    def _export(self) -> dict:
        """Get dictionary that can be used in config to duplicate this template

        Or use it as a base for another search template.
        Exported without documentation.

        Experimental, does not work with all subfeatures and may never be
        """
        request_body = pascalize(self._pass_through_defaults)
        request_body.update(
            {pascalize(k): v.template for k, v in self._templated_defaults.items()}
        )

        if "View" in request_body and isinstance(request_body["View"], search.Views):
            request_body["View"] = request_body["View"].value

        result = {"request_body": request_body}

        if self._placeholders_defaults:
            result["parameters"] = {
                name: {"default": value}
                for name, value in self._placeholders_defaults.items()
            }

        return result



"""Helper functions."""

import string
from enum import Enum
from typing import Any, Dict, List, Set, Tuple, Union


TRUTH_TEXT = frozenset(("t", "true", "y", "yes", "on", "1"))
FALSE_TEXT = frozenset(("f", "false", "n", "no", "off", "0", ""))
PROTECTED_KEYS = frozenset(("secret", "password", "passwd", "pwd", "token"))

InterpolateType = Union[bool, Dict[str, str]]


class InterpolateEnumType(Enum):
    """Interpolation Method."""

    # standard matching
    STANDARD = 0
    # interpolation will look through lower levels to attempt to resolve variables.
    # This is particularly useful for templating
    DEEP = 1
    # similar to DEEP, but interpolating will not backtrack levels.
    # That is, lower levels cannot use values from higher levels.
    DEEP_NO_BACKTRACK = 2


class AttributeDict(dict):
    """Dictionary subclass enabling attribute lookup/assignment of keys/values."""

    def __getattr__(self, key: Any) -> Any:  # noqa: D105
        try:
            return self[key]
        except KeyError:
            # to conform with __getattr__ spec
            raise AttributeError(key)

    def __setattr__(self, key: Any, value: Any) -> None:  # noqa: D105
        self[key] = value


def as_bool(s: Any) -> bool:
    """
    Boolean value from an object.

    Return the boolean value ``True`` if the case-lowered value of string
    input ``s`` is a `truthy string`. If ``s`` is already one of the
    boolean values ``True`` or ``False``, return it.
    """
    if s is None:
        return False
    if isinstance(s, bool):
        return s
    s = str(s).strip().lower()
    if s not in TRUTH_TEXT and s not in FALSE_TEXT:
        raise ValueError("Expected a valid True or False expression.")
    return s in TRUTH_TEXT


def clean(key: str, value: Any, mask: str = "******") -> Any:
    """
    Mask a value if needed.

    :param key: key
    :param value: value to hide
    :param mask: string to use in case value should be hidden
    :return: clear value or mask
    """
    key = key.lower()
    # check for protected keys
    for pk in PROTECTED_KEYS:
        if pk in key:
            return mask
    # urls
    if isinstance(value, str) and "://" in value:
        from urllib.parse import urlparse

        url = urlparse(value)
        if url.password is None:
            return value
        else:
            return url._replace(
                netloc="{}:{}@{}".format(url.username, mask, url.hostname)
            ).geturl()
    return value


def interpolate_standard(text: str, d: dict, found: Set[Tuple[str, ...]]) -> str:
    """
    Return the string interpolated as many times as needed.

    :param text: string possibly containing an interpolation pattern
    :param d: dictionary
    :param found: variables found so far
    """
    if not isinstance(text, str):
        return text

    variables = tuple(
        sorted(x[1] for x in string.Formatter().parse(text) if x[1] is not None)
    )

    if not variables:
        return text

    if variables in found:
        raise ValueError("Cycle detected while interpolating keys")
    else:
        found.add(variables)

    interpolated = {v: interpolate_standard(d[v], d, found) for v in variables}
    return text.format(**interpolated)


def interpolate_deep(
    attr: str,
    text: str,
    d: List[dict],
    resolved: Dict[str, str],
    levels: Dict[str, int],
    method: InterpolateEnumType,
) -> str:
    """
    Return the string interpolated as many times as needed.

    :param attr: attribute name
    :param text: string possibly containing an interpolation pattern
    :param d: dictionary
    :param resolved: variables resolved so far
    :param levels: last level to read the variable from
    """
    if not isinstance(text, str):
        return text

    variables = {x[1] for x in string.Formatter().parse(text) if x[1] is not None}

    if not variables:
        return text

    length = len(d)

    for variable in variables.difference(resolved.keys()):
        # start at 1 if this is the intended attribute
        level = levels.setdefault(variable, 1 if variable == attr else 0)
        # get the first level for which the variable is defined
        if level == length:
            raise KeyError(variable)
        for i, dict_ in enumerate(d[level:]):
            if variable in dict_:
                level = level + i
                break
        else:
            raise KeyError(variable)
        levels[variable] = level + 1

        new_d = (
            ([{}] * level) + d[level:]
            if method == InterpolateEnumType.DEEP_NO_BACKTRACK
            else d
        )
        resolved[variable] = interpolate_deep(
            attr, d[level][variable], new_d, resolved, levels, method
        )

    return text.format(**resolved)


def flatten(d: List[dict]) -> dict:
    """
    Flatten a list of dictionaries.

    :param d: dictionary list
    """
    result = {}
    [result.update(dict_) for dict_ in d[::-1]]
    return result


def interpolate_object(
    attr: str, obj: Any, d: List[dict], method: InterpolateEnumType
) -> Any:
    """
    Return the interpolated object.

    :param attr: attribute name
    :param obj: object to interpolate
    :param d: dictionary
    :param method: interpolation method
    """
    if isinstance(obj, str):
        if method == InterpolateEnumType.STANDARD:
            return interpolate_standard(obj, flatten(d), set())
        elif method in (
            InterpolateEnumType.DEEP,
            InterpolateEnumType.DEEP_NO_BACKTRACK,
        ):
            return interpolate_deep(attr, obj, d, {}, {}, method)
        else:
            raise ValueError('Invalid interpolation method "%s"' % method)
    elif hasattr(obj, "__iter__"):
        if isinstance(obj, tuple):
            return tuple(interpolate_object(attr, x, d, method) for x in obj)
        else:
            return [interpolate_object(attr, x, d, method) for x in obj]
    else:
        return obj



from refinitiv.data.content.ipa.curves import zc_curve_definitions


def test_zc_curves_definition_repr():
    # given
    definition = zc_curve_definitions.Definition()
    expected_value = (
        f"<refinitiv.data.content.ipa.curves.zc_curve_definitions.Definition "
        f"object at {hex(id(definition))}>"
    )

    # when
    testing_value = repr(definition)

    # then
    assert testing_value == expected_value



from datetime import datetime
import refinitiv.data as rd
import tests.integration.conftest as global_conftest


def open_desktop_session_for_streaming_chain():
    return rd.open_session(config_name=global_conftest.conf, name="desktop.my-session")


def open_deployed_session_for_streaming_chain():
    return rd.open_session(
        config_name=global_conftest.conf, name="platform.deployed-cipsnylab2"
    )


def display_add(index, constituent, streaming_chain, triggered_events_list=None):
    if triggered_events_list is not None:
        triggered_events_list.append("Add performed")
    current_time = datetime.now().time()
    print(
        f"{current_time} - Add received for {streaming_chain.name}: {index} - {constituent}"
    )


def display_remove(index, constituent, streaming_chain, triggered_events_list=None):
    if triggered_events_list is not None:
        triggered_events_list.append("Remove performed")
    current_time = datetime.now().time()
    print(
        f"{current_time} - Remove received for {streaming_chain.name}: {index} - {constituent}"
    )


def display_update(
    index, old_constituent, new_constituent, streaming_chain, triggered_events_list=None
):
    if triggered_events_list is not None:
        triggered_events_list.append("Update performed")
    current_time = datetime.now().time()
    print(
        f"{current_time} - Update received for {streaming_chain.name}: "
        f"{index} - {old_constituent} to {new_constituent}"
    )


def display_complete_decoded(constituents, streaming_chain, triggered_events_list=None):
    if triggered_events_list is not None:
        triggered_events_list.append("Complete performed")
    current_time = datetime.now().time()
    print(
        f"{current_time} - StreamingChain {streaming_chain.name} is complete decoded."
        f" - constituents = {constituents}"
    )


def display_error(error, index, streaming_chain, triggered_events_list=None):
    if triggered_events_list is not None:
        triggered_events_list.append("Error received")
    current_time = datetime.now().time()
    print(f"{current_time} - Error received for Chain: {error}")


def create_chain_stream(name):
    chain = rd.content.pricing.chain.Definition(name)
    stream = chain.get_stream()
    stream.on_error(
        lambda streaming_chain, index, error: display_error(
            streaming_chain, index, error
        )
    )
    return stream


def create_chain_stream_with_callbacks(name, triggered_events, service=None):
    definition = rd.content.pricing.chain.Definition(name, service=service)
    stream = definition.get_stream()
    stream.on_add(
        lambda index, constituent, streaming_chain: display_add(
            index, constituent, streaming_chain, triggered_events
        )
    )
    stream.on_remove(
        lambda index, constituent, streaming_chain: display_remove(
            index, constituent, streaming_chain, triggered_events
        )
    )
    stream.on_update(
        lambda index, old_constituent, new_constituent, streaming_chain: display_update(
            index,
            old_constituent,
            not new_constituent,
            streaming_chain,
            triggered_events,
        )
    )
    stream.on_complete(
        lambda constituents, streaming_chain: display_complete_decoded(
            constituents, streaming_chain, triggered_events
        )
    )
    stream.on_error(
        lambda error, index, streaming_chain: display_error(
            error, index, streaming_chain, triggered_events
        )
    )
    return stream



from refinitiv.data.content.ipa.curves._cross_currency_curves.definitions import search
from tests.unit.conftest import StubResponse, StubSession


def test_search_definition():
    # given
    response = StubResponse(
        {
            "data": [
                {
                    "curveDefinitions": [
                        {
                            "baseCurrency": "EUR",
                            "baseIndexName": "EURIBOR",
                            "name": "EUR CHF FxCross",
                            "quotedCurrency": "CHF",
                            "quotedIndexName": "LIBOR",
                            "source": "Refinitiv",
                            "isNonDeliverable": False,
                            "mainConstituentAssetClass": "Swap",
                            "riskType": "CrossCurrency",
                            "isFallbackForFxCurveDefinition": True,
                            "firstHistoricalAvailabilityDate": "2017-01-10",
                            "id": "1146241a-e32f-45f5-9309-02c959e97b96",
                        }
                    ]
                }
            ]
        }
    )
    session = StubSession(is_open=True, response=response)
    definition = search.Definition(base_currency="EUR", quoted_currency="CHF")

    # when
    response = definition.get_data(session)

    # then
    assert response.data.raw
    assert not response.data.df.empty



# coding: utf8


from enum import Enum, unique


@unique
class Format(Enum):
    """
    The enumerate that specifies whether the calculated volatilities
    """

    LIST = "List"
    MATRIX = "Matrix"
    N_DIMENSIONAL_ARRAY = "NDimensionalArray"



from typing import TYPE_CHECKING, Union

from .. import content
from .._types import OptStr, OptInt
from ..content.search import Views

if TYPE_CHECKING:
    import pandas as pd


def search(
    query: OptStr = None,
    view: Union[Views, str] = Views.SEARCH_ALL,
    filter: OptStr = None,
    order_by: OptStr = None,
    boost: OptStr = None,
    select: OptStr = None,
    top: OptInt = 10,
    skip: OptInt = 0,
    group_by: OptStr = None,
    group_count: OptInt = 3,
    features: OptStr = None,
    scope: OptStr = None,
    terms: OptStr = None,
) -> "pd.DataFrame":
    """
    This class describe parameters to retrieve data for search.

    Parameters
    ----------
    query: str, optional
        Keyword argument for view

    view: Views or str, optional
        The view for searching see at Views enum.
        Default: Views.SEARCH_ALL

    filter: str, optional
        Where query is for unstructured end-user-oriented restriction, filter is for
        structured programmatic restriction.

    order_by: str, optional
        Defines the order in which matching documents should be returned.

    boost: str, optional
        This argument supports exactly the same predicate expression syntax as filter,
        but where filter restricts which documents are matched at all,
        boost just applies a large scoring boost to documents it matches,
        which will almost always guarantee that they appear at the top of the results.

    select: str, optional
        A comma-separated list of the properties of a document to be returned in the response.

    top: int, optional
        the maximum number of documents to retrieve. Must be non-negative.
        default: 10

    skip: int, optional
        The number of documents to skip in the sorted result set before returning the
        next top.

    group_by: str, optional
        If specified, this must name a single Groupable property.
        returned documents are grouped into buckets based on their value for this
        property.

    group_count: str, optional
        When supplied in combination with group_by, sets the maximum number of documents
        to be returned per bucket.
        default: 3

    Examples
    --------
    >>> from refinitiv.data as rd
    >>> df = rd.discovery.search(query="cfo", view=rd.discovery.Views.PEOPLE)
    """
    return (
        content.search.Definition(
            query=query,
            view=view,
            filter=filter,
            order_by=order_by,
            boost=boost,
            select=select,
            top=top,
            skip=skip,
            group_by=group_by,
            group_count=group_count,
            features=features,
            scope=scope,
            terms=terms,
        )
        .get_data()
        .data.df
    )



import allure
import pytest

from refinitiv.data.content.esg import basic_overview
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ownership.helpers import (
    check_response_dataframe_instrument_contains_expected_universes,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_response_status,
    check_non_empty_response_data,
    check_response_dataframe_contains_columns_names,
    check_dataframe_column_date_for_datetime_type,
)

EXPECTED_TITLES = [
    "Instrument",
    "Period End Date",
    "ESG Reporting Scope",
    "ESG Period Last Update Date",
    "CO2 Equivalents Emission Total",
    "Women Managers",
    "Average Training Hours",
]


@allure.suite("Content object - ESG Basic Overview")
@allure.feature("Content object - ESG Basic Overview")
@allure.severity(allure.severity_level.CRITICAL)
class TestESGBasicOverview:
    @allure.title(
        "Create ESG Basic Overview definition object with valid params - synchronously"
    )
    @pytest.mark.parametrize(
        "universe",
        ["IBM.N", ["IBM.N", "VOD.L"]],
    )
    @pytest.mark.caseid("C40203262")
    @pytest.mark.smoke
    def test_esg_basic_overview_definition_object_with_valid_params_and_get_data(
        self, universe, open_platform_session
    ):
        response = basic_overview.Definition(universe=universe).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_response_dataframe_contains_columns_names(response, EXPECTED_TITLES)
        check_response_dataframe_instrument_contains_expected_universes(
            response, universe
        )
        check_non_empty_response_data(response)
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title(
        "Create ESG Basic Overview definition object with invalid params and get error"
    )
    @pytest.mark.caseid("C40203263")
    @pytest.mark.parametrize(
        "universe,expected_error",
        [
            (
                [],
                "Error code 400 | Validation error: Missing required parameter 'universe'",
            ),
            (
                ["INVALID"],
                "Error code 412 | Unable to resolve all requested identifiers. Requested items: ['INVALID']",
            ),
        ],
    )
    def test_esg_basic_overview_definition_object_with_invalid_params_and_get_error(
        self,
        universe,
        expected_error,
        open_platform_session,
    ):
        with pytest.raises(RDError) as error:
            basic_overview.Definition(universe=universe).get_data()
        assert str(error.value) == expected_error

    @allure.title("Create ESG Basic Overview definition object using closed session")
    @pytest.mark.caseid("C40203264")
    def test_esg_basic_overview_definition_object_using_closed_session(
        self, open_platform_session
    ):
        session = open_platform_session
        session.close()
        with pytest.raises(ValueError) as error:
            basic_overview.Definition(universe="IBM.N").get_data()
        assert str(error.value) == "Session is not opened. Can't send any request"

    @allure.title(
        "Create ESG Basic Overview definition object with use mix valid and invalid data - asynchronously"
    )
    @pytest.mark.caseid("C40203265")
    @pytest.mark.parametrize(
        "universe,invalid_universe", [(["IBM.N", "INVALID"], "INVALID")]
    )
    async def test_esg_basic_overview_definition_object_with_use_mix_valid_and_invalid_data(
        self, open_platform_session_async, universe, invalid_universe
    ):
        valid_response, invalid_response = await get_async_response_from_definitions(
            basic_overview.Definition(universe=universe),
            basic_overview.Definition(universe=invalid_universe),
        )
        check_response_status(
            response=valid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code=None,
            expected_error_message="Failed to resolve identifiers ['INVALID']",
        )
        check_non_empty_response_data(valid_response)
        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code=HttpStatusCode.FOUR_HUNDRED_TWELVE,
            expected_error_message="Unable to resolve all requested identifiers. Requested items: ['INVALID']",
        )



# coding: utf8


from ._surface_request_item import SurfaceRequestItem
from ._enums import UnderlyingType
from ._i_ir_vol_model_definition import IIrVolModelDefinition as CapSurfaceDefinition
from ._i_ir_vol_model_pricing_parameters import (
    IIrVolModelPricingParameters as CapCalculationParams,
)


class CapSurfaceRequestItem(SurfaceRequestItem):
    def __init__(
        self,
        instrument_type,
        surface_layout,
        surface_params,
        underlying_definition,
        surface_tag,
    ):
        super().__init__(
            surface_layout=surface_layout,
            surface_tag=surface_tag,
            underlying_type=UnderlyingType.CAP,
        )
        self.instrument_type = instrument_type
        self.surface_parameters = surface_params
        self.underlying_definition = underlying_definition

    @property
    def surface_parameters(self):
        """
        :return: object CapCalculationParams
        """
        return self._get_object_parameter(CapCalculationParams, "surfaceParameters")

    @surface_parameters.setter
    def surface_parameters(self, value):
        self._set_object_parameter(CapCalculationParams, "surfaceParameters", value)

    @property
    def underlying_definition(self):
        """
        :return: object CapSurfaceDefinition
        """
        return self._get_object_parameter(CapSurfaceDefinition, "underlyingDefinition")

    @underlying_definition.setter
    def underlying_definition(self, value):
        self._set_object_parameter(CapSurfaceDefinition, "underlyingDefinition", value)

    @property
    def instrument_type(self):
        """
        The type of instrument being defined.
        :return: str
        """
        return self._get_parameter("instrumentType")

    @instrument_type.setter
    def instrument_type(self, value):
        self._set_parameter("instrumentType", value)



from typing import TYPE_CHECKING

from ._fx_forward_curve_definition import FxForwardCurveDefinition
from ._fx_forward_curve_parameters import FxForwardCurveParameters
from ._fx_shift_scenario import FxShiftScenario
from ...._object_definition import ObjectDefinition
from ......_tools import try_copy_to_list

if TYPE_CHECKING:
    from ......_types import OptStr
    from ._types import CurveDefinition, CurveParameters, ShiftScenarios


class RequestItem(ObjectDefinition):
    """
    Generates the Cross Currency curves for the definitions provided

    Parameters
    ----------

    curve_definition : FxForwardCurveDefinition, optional

    curve_parameters : FxForwardCurveParameters, optional

    shift_scenarios : list of FxShiftScenario, optional
        The list of attributes applied to the curve shift scenarios.
    curve_tag : str, optional
        A user-defined string to identify the interest rate curve. it can be used to
        link output results to the curve definition. limited to 40 characters. only
        alphabetic, numeric and '- _.#=@' characters are supported.
    """

    def __init__(
        self,
        curve_definition: "CurveDefinition" = None,
        curve_parameters: "CurveParameters" = None,
        shift_scenarios: "ShiftScenarios" = None,
        curve_tag: "OptStr" = None,
    ) -> None:
        super().__init__()
        self.curve_definition = curve_definition
        self.curve_parameters = curve_parameters
        self.shift_scenarios = try_copy_to_list(shift_scenarios)
        self.curve_tag = curve_tag

    @property
    def curve_definition(self):
        """
        :return: object FxForwardCurveDefinition
        """
        return self._get_object_parameter(FxForwardCurveDefinition, "curveDefinition")

    @curve_definition.setter
    def curve_definition(self, value):
        self._set_object_parameter(FxForwardCurveDefinition, "curveDefinition", value)

    @property
    def curve_parameters(self):
        """
        :return: object FxForwardCurveParameters
        """
        return self._get_object_parameter(FxForwardCurveParameters, "curveParameters")

    @curve_parameters.setter
    def curve_parameters(self, value):
        self._set_object_parameter(FxForwardCurveParameters, "curveParameters", value)

    @property
    def shift_scenarios(self):
        """
        The list of attributes applied to the curve shift scenarios.
        :return: list FxShiftScenario
        """
        return self._get_list_parameter(FxShiftScenario, "shiftScenarios")

    @shift_scenarios.setter
    def shift_scenarios(self, value):
        self._set_list_parameter(FxShiftScenario, "shiftScenarios", value)

    @property
    def curve_tag(self):
        """
        A user-defined string to identify the interest rate curve. it can be used to
        link output results to the curve definition. limited to 40 characters. only
        alphabetic, numeric and '- _.#=@' characters are supported.
        :return: str
        """
        return self._get_parameter("curveTag")

    @curve_tag.setter
    def curve_tag(self, value):
        self._set_parameter("curveTag", value)



from typing import TYPE_CHECKING, List

from .._content_data_validator import ContentDataValidator
from .._request_factory import DatesAndCalendarsResponseFactory
from ...._content_data import Data
from ...._content_data_provider import ContentDataProvider
from ...._df_builder import default_build_df
from .....content.ipa._content_provider import DatesAndCalendarsRequestFactory
from .....delivery._data._data_provider import ValidatorContainer

if TYPE_CHECKING:
    from .....delivery._data._data_provider import ParsedData


class Period:
    def __init__(self, count: float, tenor: str, tag: str = ""):
        self._count = count
        self._tenor = tenor
        self._tag = tag

    @property
    def count(self):
        return self._count

    @property
    def tenor(self):
        return self._tenor

    @property
    def tag(self):
        return self._tag


class CountedPeriods(Data):
    _counted_periods = None

    @property
    def counted_periods(self):
        if self._counted_periods is None:
            self._counted_periods = [
                Period(item["count"], item["tenor"], tag=item.get("tag"))
                for item in self._raw
            ]

        return self._counted_periods

    def __getitem__(self, item: int):
        return self.counted_periods[item]


class CountedPeriod(Data):
    def __init__(self, raw: dict, count: float, tenor: str, tag: str = "", **kwargs):
        super().__init__(raw, **kwargs)
        self._period = Period(count, tenor, tag)

    @property
    def counted_period(self):
        return self._period


class CountPeriodsResponseFactory(DatesAndCalendarsResponseFactory):
    def create_data_success(self, parsed_data: "ParsedData", **kwargs):
        raw: List[dict] = self.get_raw(parsed_data)
        for item in raw:
            if item.get("error"):
                return self.create_fail(item, **kwargs)

        if len(raw) > 1:
            data = CountedPeriods(raw, default_build_df)

        else:
            raw_item = raw[0]
            data = CountedPeriod(
                raw,
                raw_item["count"],
                raw_item["tenor"],
                raw_item.get("tag"),
                dfbuilder=default_build_df,
                **kwargs,
            )

        return data


count_periods_data_provider = ContentDataProvider(
    request=DatesAndCalendarsRequestFactory(),
    response=CountPeriodsResponseFactory(),
    validator=ValidatorContainer(content_validator=ContentDataValidator()),
)



from ...._curves._cross_currency_curves._enums import (
    InterpolationMode,
    MainConstituentAssetClass,
    RiskType,
)

from ...._curves._cross_currency_curves._curves._enums import (
    ArrayMainConstituentAssetClass,
    ArrayRiskType,
    ConstituentOverrideMode,
    ShiftType,
    ShiftUnit,
)



from inspect import signature

import pytest

import refinitiv.data.content.ipa.financial_contracts as rdf
from refinitiv.data.delivery._data._data_provider_factory import make_provider
from tests.unit.conftest import StubSession
from tests.unit.conftest import (
    remove_dunder_methods,
    remove_private_attributes,
    has_property_names_in_class,
    get_property_names,
)


def test_ipa_swaption_definition_attributes():
    expected_attributes = ["get_data", "get_data_async", "get_stream"]
    testing_attributes = dir(rdf.swaption.Definition)
    testing_attributes = remove_dunder_methods(testing_attributes)
    testing_attributes = remove_private_attributes(testing_attributes)
    assert expected_attributes == testing_attributes


def test_ipa_swaption_attributes():
    expected_attributes = [
        "BermudanSwaptionDefinition",
        "BuySell",
        "CallPut",
        "Definition",
        "ExerciseScheduleType",
        "ExerciseStyle",
        "InputFlow",
        "PremiumSettlementType",
        "PriceSide",
        "PricingParameters",
        "SwaptionMarketDataRule",
        "SwaptionSettlementType",
        "SwaptionType",
    ]
    testing_attributes = dir(rdf.swaption)
    testing_attributes = remove_dunder_methods(testing_attributes)
    testing_attributes = remove_private_attributes(testing_attributes)
    assert expected_attributes == testing_attributes, testing_attributes


@pytest.mark.parametrize(
    argnames="input_data",
    ids=[
        "Definition",
        "PricingParameters",
        "SwaptionMarketDataRule",
        "BermudanSwaptionDefinition",
        "InputFlow",
    ],
    argvalues=[
        (
            rdf.swaption._swaption_definition.SwaptionInstrumentDefinition,
            {
                "notional_amount": 12.321,
                "spread_vs_atm_in_bp": 123.321,
                "instrument_tag": "instrument_tag",
                "end_date": "end_date",
                "tenor": "tenor",
                "bermudan_swaption_definition": rdf.swaption.BermudanSwaptionDefinition(),
                "buy_sell": rdf.swaption.BuySell.BUY,
                "exercise_style": rdf.swaption.ExerciseStyle.EURO,
                "settlement_type": rdf.swaption.SwaptionSettlementType.CASH,
                "underlying_definition": rdf.swap.Definition(
                    tenor="5Y", template="template"
                ),
                "strike_percent": 10.10,
                "swaption_type": rdf.swaption.SwaptionType.PAYER,
                "start_date": "start_date",
                "premium_settlement_type": rdf.swaption.PremiumSettlementType.FORWARD,
                "payments": [rdf.swaption.InputFlow()],
            },
        ),
        (
            rdf.swaption.PricingParameters,
            {
                "market_data_date": "market_data_date",
                "market_value_in_deal_ccy": 3.3,
                "nb_iterations": 4,
                "valuation_date": "valuation_date",
                "price_side": rdf.swaption.PriceSide.ASK,
                "simulate_exercise": True,
                "exercise_date": "exercise_date",
                "report_ccy": "report_ccy",
            },
        ),
        (
            rdf.swaption.SwaptionMarketDataRule,
            {
                "discount": "discount",
                "forward": "forward",
            },
        ),
        (
            rdf.swaption.BermudanSwaptionDefinition,
            {
                "exercise_schedule": ["exercise_schedule"],
                "exercise_schedule_type": rdf.swaption.ExerciseScheduleType.FIXED_LEG,
                "notification_days": 3,
            },
        ),
        (
            rdf.swaption.InputFlow,
            {
                "amount": 1.1,
                "premium_settlement_type": rdf.swaption.PremiumSettlementType.SPOT,
                "currency": "currency",
                "date": "date",
            },
        ),
    ],
)
def test_parameter(input_data):
    cls, kwargs = input_data
    args_names = list(kwargs.keys())
    inst = cls(**kwargs)

    s = signature(cls.__init__)
    # +1 for (self)
    assert len(s.parameters) == (
        len(args_names) + 1
    ), f"csl={cls}{str(set(get_property_names(cls)) - set(args_names))}"

    assert has_property_names_in_class(cls, args_names), set(args_names) - set(
        get_property_names(cls)
    )

    for k, v in kwargs.items():
        attr = getattr(inst, k)
        assert attr == v, k


def test_request_body():
    # given
    expected_body = {
        "fields": ["1", "2"],
        "universe": [
            {
                "instrumentDefinition": {
                    "bermudanSwaptionDefinition": {
                        "exerciseSchedule": ["exercise_schedule"],
                        "notificationDays": 0,
                    },
                    "startDate": "start_date",
                    "underlyingDefinition": {"tradeDate": "trade_date"},
                },
                "extended_params": "extended_params",
                "instrumentType": "Swaption",
                "pricingParameters": {"marketDataDate": "market_data_date"},
            }
        ],
    }
    definition = rdf.swaption.Definition(
        bermudan_swaption_definition=rdf.swaption.BermudanSwaptionDefinition(
            exercise_schedule=["exercise_schedule"],
            notification_days=0,
        ),
        start_date="start_date",
        fields=["1", "2"],
        underlying_definition=rdf.swap.Definition(trade_date="trade_date"),
        pricing_parameters=rdf.swaption.PricingParameters(
            market_data_date="market_data_date"
        ),
        extended_params={"extended_params": "extended_params"},
    )

    # when
    content_type = definition._kwargs["__content_type__"]
    provider = make_provider(content_type)
    request = provider.request.create(StubSession(), "url", **definition._kwargs)

    # then
    testing_body = request.json
    assert testing_body == expected_body, expected_body



import time

import allure
import pytest

import refinitiv.data as rd
from refinitiv.data.content import trade_data_service
from tests.integration.content.trade_data_service.conftest import (
    create_trade_data_service_with_all_callbacks,
    check_triggered_event_list,
)

ONE_SEC = 1


@allure.suite("Content object - Trade Data Service")
@allure.feature("Content object - Trade Data Service")
@allure.severity(allure.severity_level.CRITICAL)
@pytest.mark.underlying_content("tds")
class TestTradeDataService:
    @allure.title("Create TDS object when session is not open")
    @pytest.mark.parametrize("universe", [["U1"]])
    @pytest.mark.caseid("33468307")
    def test_create_tds_when_session_is_not_open(self, universe):
        with pytest.raises(
            AttributeError,
            match="No default session created yet. Please create a session first!",
        ):
            rd.session.set_default(None)
            stream = create_trade_data_service_with_all_callbacks(universe=universe)
            stream.open()

    @allure.title("Create TDS object in session with default values")
    @pytest.mark.parametrize("universe", [["U1"]])
    @pytest.mark.caseid("33466883")
    def test_open_trade_data_service_with_default_values(self, universe, open_session):
        triggered_events_list = []
        stream = create_trade_data_service_with_all_callbacks(
            universe=universe,
            api="streaming.trading-analytics.endpoints.redi",
            session=open_session,
            triggered_events_list=triggered_events_list,
        )
        stream.open()
        time.sleep(ONE_SEC)
        stream.close()
        check_triggered_event_list(triggered_events_list)

    @allure.title(
        "Create TDS object in session with a specific universe name and without specifying fields"
    )
    @pytest.mark.parametrize("universe", [["U1"]])
    @pytest.mark.caseid("33466895")
    @pytest.mark.smoke
    def test_create_tds_object_by_specific_universe(self, universe, open_session):
        triggered_events_list = []
        stream = create_trade_data_service_with_all_callbacks(
            universe=universe,
            session=open_session,
            triggered_events_list=triggered_events_list,
        )
        stream.open()

        actual_fields_list = stream._stream._headers_ids
        actual_universe = stream._stream._universe

        assert (
            len(actual_fields_list) == 57
        ), f"Actual fields length = {len(actual_fields_list)}"
        assert (
            actual_universe == universe
        ), f"Universe not found, {actual_universe} != {universe}"

    @allure.title("Create TDS object in session before closing session")
    @pytest.mark.parametrize("universe", [["U1"]])
    @pytest.mark.caseid("33466895")
    def test_create_tds_in_session_before_closing_session(self, universe, open_session):
        with pytest.raises(AssertionError, match="Session must be open"):
            session = open_session
            stream = create_trade_data_service_with_all_callbacks(
                universe=universe,
                session=open_session,
            )
            session.close()
            stream.open()

    @allure.title(
        "Create TDS object in session with default values and without any parameters"
    )
    @pytest.mark.caseid("C33466888")
    def test_open_trade_data_service_without_any_params(self, open_platform_session):
        stream = trade_data_service.Definition().get_stream(
            session=open_platform_session
        )
        stream.open()
        time.sleep(ONE_SEC)
        assert stream._stream.state.name == "Opened"

    @allure.title(
        "Stop listening closed TDS stream on session and keep retrieving data for active stream"
    )
    @pytest.mark.parametrize("universe", [["U2"]])
    @pytest.mark.caseid("33468283")
    def test_keep_retrieving_data_for_active_stream(self, universe, open_session):
        triggered_events_list_01 = []
        stream_01 = create_trade_data_service_with_all_callbacks(
            universe=universe,
            universe_type=trade_data_service.UniverseTypes.RIC,
            session=open_session,
            triggered_events_list=triggered_events_list_01,
        )
        stream_01.open()

        triggered_events_list_02 = []
        stream_02 = create_trade_data_service_with_all_callbacks(
            universe=universe,
            universe_type=trade_data_service.UniverseTypes.UserID,
            session=open_session,
            triggered_events_list=triggered_events_list_02,
        )
        stream_02.open()
        stream_01.close()
        time.sleep(ONE_SEC)
        check_triggered_event_list(triggered_events_list_02)
        stream_02.close()

    @allure.title(
        "Receiving TDS data when the session was interrupted during the process"
    )
    @pytest.mark.parametrize("universe", [[]])
    @pytest.mark.caseid("33468321")
    def test_receiving_tds_when_session_was_interrupted(self, universe, open_session):
        with pytest.raises(AssertionError, match="Session must be open"):
            session = open_session
            stream = trade_data_service.Definition(
                universe=universe,
                finalized_orders=trade_data_service.FinalizedOrders.P1D,
                events=trade_data_service.Events.Full,
            ).get_stream(session=open_session)
            stream.open()
            session.close()
            stream.open()

    @allure.title(
        "Create TDS object in session with a specific universe name, fields and extended params"
    )
    @pytest.mark.parametrize(
        "universe,extended_fields",
        [(["U1"], ["InstrumentLongName", "OrderLifeDuration"])],
    )
    @pytest.mark.caseid("33468337")
    def test_create_trade_data_service_with_extended_params(
        self, universe, extended_fields, open_session
    ):
        stream = create_trade_data_service_with_all_callbacks(
            universe=universe,
            fields=["PrimaryListedRIC"],
            session=open_session,
            universe_type=trade_data_service.UniverseTypes.UserID,
            extended_params={"universe": ["U2"], "view": extended_fields},
        )
        stream.open()
        time.sleep(ONE_SEC)

        actual_fields_list = stream._stream._headers_ids

        assert (
            len(actual_fields_list) == 3
        ), f"Actual fields length = {len(actual_fields_list)}"
        for fields in extended_fields:
            assert fields in actual_fields_list, f"{fields} not in {actual_fields_list}"

        stream.close()

    @allure.title(
        "Checking the receipt of data using the user type RIC, UserID or Symbol"
    )
    @pytest.mark.parametrize(
        "universe,universe_type",
        [
            (["INUV"], trade_data_service.UniverseTypes.RIC),
            (["U10"], trade_data_service.UniverseTypes.UserID),
            (["MARA"], trade_data_service.UniverseTypes.Symbol),
        ],
    )
    @pytest.mark.caseid("")
    def test_checking_the_receipt_of_data_using_the_user_type_ric_or_symbol(
        self, universe, universe_type, open_session
    ):
        triggered_events_list = []
        stream = create_trade_data_service_with_all_callbacks(
            universe=universe,
            fields=["OrderKey", "InstrumentId", "PrimaryListedRIC", "CompositeRIC"],
            session=open_session,
            universe_type=universe_type,
            triggered_events_list=triggered_events_list,
        )
        stream.open()
        time.sleep(ONE_SEC)
        stream.close()
        check_triggered_event_list(triggered_events_list)

    @allure.title("Checking that no data was received when using an invalid user type")
    @pytest.mark.parametrize(
        "universe,universe_type",
        [
            (["AAPL.OQ"], trade_data_service.UniverseTypes.Symbol),
            # (["AAPL"], trade_data_service.UniverseTypes.UserID),
            (["U1"], trade_data_service.UniverseTypes.Symbol),
            (["U1"], trade_data_service.UniverseTypes.RIC),
        ],
    )
    @pytest.mark.caseid("")
    def test_checking_that_no_data_was_received(
        self, universe, universe_type, open_session
    ):
        triggered_events_list = []
        stream = create_trade_data_service_with_all_callbacks(
            universe=universe,
            session=open_session,
            universe_type=universe_type,
            triggered_events_list=triggered_events_list,
        )
        stream.open()
        time.sleep(ONE_SEC)
        stream.close()
        check_triggered_event_list(triggered_events_list)

    @pytest.mark.caseid("")
    @allure.title("Create trade data service stream with invalid api parameter")
    @pytest.mark.parametrize(
        "universe,universe_type",
        [
            (["U1"], trade_data_service.UniverseTypes.Symbol),
        ],
    )
    def test_create_trade_data_service_with_invalid_api(
        self, open_desktop_session, universe, universe_type
    ):
        with pytest.raises(ValueError) as error:
            stream = create_trade_data_service_with_all_callbacks(
                api="streaming.trading-analytics.endpoints.invalid",
                universe=universe,
                session=open_desktop_session,
                universe_type=universe_type,
            )
            stream.open()
        assert (
            str(error.value)
            == "Not an existing path apis.streaming.trading-analytics.endpoints.invalid to url into config file"
        )



__all__ = (
    "AdjustInterestToPaymentDate",
    "AmortizationFrequency",
    "AmortizationItem",
    "AmortizationType",
    "BusinessDayConvention",
    "DateRollingConvention",
    "DayCountBasis",
    "Definition",
    "Direction",
    "Frequency",
    "IndexAverageMethod",
    "IndexCompoundingMethod",
    "IndexConvexityAdjustmentIntegrationMethod",
    "IndexConvexityAdjustmentMethod",
    "IndexObservationMethod",
    "IndexResetType",
    "IndexSpreadCompoundingMethod",
    "InputFlow",
    "InterestCalculationConvention",
    "InterestType",
    "LegDefinition",
    "NotionalExchange",
    "PremiumSettlementType",
    "PriceSide",
    "PricingParameters",
    "StubRule",
    "SwaptionType",
    "TenorReferenceDate",
)

from ._definition import Definition
from ._enums import (
    AdjustInterestToPaymentDate,
    AmortizationFrequency,
    AmortizationType,
    BusinessDayConvention,
    DateRollingConvention,
    DayCountBasis,
    Direction,
    Frequency,
    IndexAverageMethod,
    IndexCompoundingMethod,
    IndexConvexityAdjustmentIntegrationMethod,
    IndexConvexityAdjustmentMethod,
    IndexObservationMethod,
    IndexResetType,
    IndexSpreadCompoundingMethod,
    InterestCalculationConvention,
    InterestType,
    NotionalExchange,
    PremiumSettlementType,
    PriceSide,
    StubRule,
    SwaptionType,
    TenorReferenceDate,
)
from ._models import AmortizationItem, InputFlow
from ._swap_leg_definition import LegDefinition
from ._swap_pricing_parameters import PricingParameters



from typing import Optional, TYPE_CHECKING

from ._bid_ask_fields_description import BidAskFieldsDescription
from ._formula_parameter_description import FormulaParameterDescription
from ._fx_forward_instrument_definition import FxForwardInstrumentDefinition
from ...._object_definition import ObjectDefinition


if TYPE_CHECKING:
    from ......_types import OptStr


class FxForwardInstrumentDescription(ObjectDefinition):
    """
    Creates the Cross Currency curve definition with the definition provided.

    Parameters
    ----------
    fields : BidAskFieldsDescription, optional

    formula_parameters : FormulaParameterDescription, optional

    instrument_definition : FxForwardInstrumentDefinition, optional

    formula : str, optional

    """

    def __init__(
        self,
        fields: Optional[BidAskFieldsDescription] = None,
        formula_parameters: Optional[FormulaParameterDescription] = None,
        instrument_definition: Optional[FxForwardInstrumentDefinition] = None,
        formula: "OptStr" = None,
    ) -> None:
        super().__init__()
        self.fields = fields
        self.formula_parameters = formula_parameters
        self.instrument_definition = instrument_definition
        self.formula = formula

    @property
    def fields(self):
        """
        :return: object BidAskFieldsDescription
        """
        return self._get_object_parameter(BidAskFieldsDescription, "fields")

    @fields.setter
    def fields(self, value):
        self._set_object_parameter(BidAskFieldsDescription, "fields", value)

    @property
    def formula_parameters(self):
        """
        :return: list FormulaParameterDescription
        """
        return self._get_list_parameter(
            FormulaParameterDescription, "formulaParameters"
        )

    @formula_parameters.setter
    def formula_parameters(self, value):
        self._set_list_parameter(
            FormulaParameterDescription, "formulaParameters", value
        )

    @property
    def instrument_definition(self):
        """
        :return: object FxForwardInstrumentDefinition
        """
        return self._get_object_parameter(
            FxForwardInstrumentDefinition, "instrumentDefinition"
        )

    @instrument_definition.setter
    def instrument_definition(self, value):
        self._set_object_parameter(
            FxForwardInstrumentDefinition, "instrumentDefinition", value
        )

    @property
    def formula(self):
        """
        :return: str
        """
        return self._get_parameter("formula")

    @formula.setter
    def formula(self, value):
        self._set_parameter("formula", value)
